import os
from typing import Optional, Iterable, Union
from pydantic import BaseModel
from icolos.core.workflow_steps.step import StepBase
from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum
from icolos.utils.execute_external.license_token_guard import (
    TokenGuardParameters,
    SchrodingerLicenseTokenGuard,
)
from icolos.utils.general.files_paths import attach_root_path
from icolos.utils.execute_external.schrodinger import SchrodingerExecutor
from icolos.utils.enums.step_enums import StepDesmondEnum
from icolos.core.workflow_steps.step import _LE
import re
from shutil import copy
from typing import Dict


_EE = SchrodingerExecutablesEnum()
_SDE = StepDesmondEnum()


class StepSchrodingerBase(StepBase, BaseModel):

    token_guard: Optional[TokenGuardParameters] = None

    def __init__(self, **data):
        super().__init__(**data)

    def _apply_token_guard(self):
        if self.token_guard is not None:
            token_guard = SchrodingerLicenseTokenGuard(token_guard=self.token_guard)
            token_guard.guard()

    # TODO: Deprecated - use self.converter
    def _translate_SDF_to_MAE(
        self, sdf_path: str, mae_path: str, executor: SchrodingerExecutor
    ):
        """As "Glide" is only able to read MAE (Maestro) files, write the ligands out in that format."""

        # call "sdconvert" from Schrodinger's software
        arguments = [
            "".join([_EE.SDCONVERT_I, _EE.SDCONVERT_FORMAT_SD]),
            sdf_path,
            "".join([_EE.SDCONVERT_O, _EE.SDCONVERT_FORMAT_MAE]),
            mae_path,
        ]
        executor.execute(command=_EE.SDCONVERT, arguments=arguments, check=True)

    def _translate_MAE_to_SDF(
        self, mae_path: str, sdf_path: str, executor: SchrodingerExecutor
    ):
        """In cases where the write-out mode for Glide is not producing SDF files."""

        # call "sdconvert" from Schrodinger's software
        arguments = [
            "".join([_EE.SDCONVERT_I, _EE.SDCONVERT_FORMAT_MAE]),
            mae_path,
            "".join([_EE.SDCONVERT_O, _EE.SDCONVERT_FORMAT_SD]),
            sdf_path,
        ]
        executor.execute(command=_EE.SDCONVERT, arguments=arguments, check=True)

    def _translate_PDB_to_MAE(
        self, pdb_path: str, mae_path: str, executor: SchrodingerExecutor
    ):
        """In cases where the write-out mode for Glide is not producing SDF files."""

        # call "sdconvert" from Schrodinger's software
        arguments = [
            "".join([_EE.SDCONVERT_I, _EE.STRUCTCAT_FORMAT_PDB]),
            pdb_path,
            "".join([_EE.SDCONVERT_O, _EE.SDCONVERT_FORMAT_MAE]),
            mae_path,
        ]
        executor.execute(command=_EE.STRUCTCONVERT, arguments=arguments, check=True)

    def _replace_config_value(self, key, value, config):
        value = str(value)
        pattern = rf"({key} =).*"
        pattern = re.compile(pattern)
        config = re.sub(pattern, rf"\1 {value}", config)
        return config

    def _get_template(self, file_name):
        file = [
            file
            for file in os.listdir(attach_root_path("icolos/config/desmond"))
            if file_name in file
        ]
        assert len(file) == 1
        return file[0]

    def _write_config(self, tmp_dir, dict_: Dict, file_name):
        # see if a config file was specified, assume no further changes:
        if _SDE.CONFIG in dict_.keys() and dict_[_SDE.CONFIG] is not None:
            copy(dict_[_SDE.CONFIG], tmp_dir)
        else:
            template = self._get_template(file_name)
            with open(attach_root_path(f"icolos/config/desmond/{template}"), "r") as f:
                config = f.read()
            for k, v in dict_.items():
                config = self._replace_config_value(k, v, config)

            self._logger.log(f"Compiled  file {file_name}...", _LE.DEBUG)
            for line in config.split("\n"):
                self._logger_blank.log(line, _LE.DEBUG)
            with open(os.path.join(tmp_dir, file_name), "w") as f:
                f.write(config)

    def _parse_arguments(self, defaults: Dict = {}):
        args = []

        for flag in self.settings.arguments.flags:
            args.append(flag)
        if "-WAIT" not in args:
            args.append("-WAIT")
        for k, v in self.settings.arguments.parameters.items():
            args.append(k)
            args.append(v)
        for k, v in defaults.items():
            if k not in args:
                args.append(k)
                args.append(v)
        return args

    @staticmethod
    def _parse_maestro_in_file(
        lines: Iterable[str],
    ) -> Dict[str, Union[str, Dict[str, str]]]:
        """Parses Maestro input, and returns keywords dict for it."""

        separator3 = "   "
        indent4 = "    "
        block_starters = {
            "[CONSTRAINT_GROUP",
            "[FEATURE",
        }

        # All Glide keywords. Get all keywords with:
        #   $ module load schrodinger
        #   $ glide -docking-keywords | cut -d' ' -f1 | sed 's/.*/"&"/' | paste -sd , -
        # List keywords, get first word, wrap in quotes, join lines.
        # See:
        #   - https://stackoverflow.com/a/19145499
        #   - https://unix.stackexchange.com/a/251362
        allowed_keywords = {
            "AMIDE_MODE",
            "AMIDE_TRANS_ALL",
            "AMIDE_TRANSTOL",
            "ASL_RES_INTERACTION",
            "CALC_INPUT_RMS",
            "CANONICALIZE",
            "COMPRESS_POSES",
            "CORE_ATOMS",
            "CORE_DEFINITION",
            "CORE_FILTER",
            "CORE_POS_MAX_RMSD",
            "CORE_RESTRAIN",
            "CORE_RESTRAIN_V",
            "CORE_SMARTS",
            "CORE_SNAP",
            "CORECONS_FALLBACK",
            "CSV_PROPS_FILE",
            "CV_CUTOFF",
            "DIELMOD",
            "DOCKING_METHOD",
            "DOINTRA",
            "DOINTRA_SCALE",
            "DSCORE_CUTOFF",
            "EPIK_PENALTIES",
            "EXPANDED_SAMPLING",
            "FITDEN",
            "FORCEFIELD",
            "FORCEPLANAR",
            "GLIDE_CONFGEN_BADDIST2",
            "GLIDE_CONFGEN_EFCUT",
            "GLIDE_CONS_FEAT_FILE",
            "GLIDE_CONS_FINALONLY",
            "GLIDE_CONS_RMETCOORD",
            "GLIDE_CONS_RNOEMAX",
            "GLIDE_CONS_RNOEMIN",
            "GLIDE_CONS_RPOS",
            "GLIDE_CONS_XMETCOORD",
            "GLIDE_CONS_XNOE",
            "GLIDE_CONS_XPOS",
            "GLIDE_CONS_YMETCOORD",
            "GLIDE_CONS_YNOE",
            "GLIDE_CONS_YPOS",
            "GLIDE_CONS_ZMETCOORD",
            "GLIDE_CONS_ZNOE",
            "GLIDE_CONS_ZPOS",
            "GLIDE_DIELCO",
            "GLIDE_ELEMENTS",
            "GLIDE_EXVOL_PENAL_NUM",
            "GLIDE_EXVOL_PENAL_STRENGTH",
            "GLIDE_NTOTALCONS",
            "GLIDE_NUMEXVOL",
            "GLIDE_NUMMETCOORDCONS",
            "GLIDE_NUMMETCOORDSITES",
            "GLIDE_NUMNOECONS",
            "GLIDE_NUMPOSITCONS",
            "GLIDE_NUMUSEXVOL",
            "GLIDE_OUTPUT_USEHTOR",
            "GLIDE_REFLIG_FORMAT",
            "GLIDE_REXVOL",
            "GLIDE_REXVOLIN",
            "GLIDE_TORCONS_ALLBONDS",
            "GLIDE_TORCONS_IATOMS",
            "GLIDE_TORCONS_JATOMS",
            "GLIDE_TORCONS_KATOMS",
            "GLIDE_TORCONS_LATOMS",
            "GLIDE_TORCONS_PATTERN_INDEX",
            "GLIDE_TORCONS_PATTERNS",
            "GLIDE_TORCONS_SETVAL",
            "GLIDE_TORCONS_VALUES",
            "GLIDE_TORCONSFILE",
            "GLIDE_XEXVOL",
            "GLIDE_XP_NMAXCORE",
            "GLIDE_XP_RMSCUT",
            "GLIDE_YEXVOL",
            "GLIDE_ZEXVOL",
            "GLIDECONS",
            "GLIDECONSFEATATOMS",
            "GLIDECONSFEATHASINCLUDE",
            "GLIDECONSFEATINCLUDE",
            "GLIDECONSFEATINDEX",
            "GLIDECONSFEATPATTERNS",
            "GLIDECONSGROUPNREQUIRED",
            "GLIDECONSNAMES",
            "GLIDECONSUSEMET",
            "GLIDESCORUSEMET",
            "GLIDEUSEALLEXVOL",
            "GLIDEUSECONSFEAT",
            "GLIDEUSECONSFEATINDEX",
            "GLIDEUSECONSGROUPINDEX",
            "GLIDEUSECONSLABELS",
            "GLIDEUSEXVOL",
            "GLIDEUSEXVOLNAMES",
            "GLIDEXVOLNAMES",
            "GRIDFILE",
            "GSCORE",
            "GSCORE_CUTOFF",
            "HAVEGLIDECONSFEAT",
            "HBOND_ACCEP_HALO",
            "HBOND_CUTOFF",
            "HBOND_DONOR_AROMH",
            "HBOND_DONOR_AROMH_CHARGE",
            "HBOND_DONOR_HALO",
            "INCLUDE_INPUT_CONF",
            "INCLUDE_INPUT_RINGS",
            "JOBNAME",
            "KEEP_SUBJOB_POSES",
            "KEEPRAW",
            "KEEPSKIPPED",
            "LIG_CCUT",
            "LIG_MAECHARGES",
            "LIG_VSCALE",
            "LIGAND_END",
            "LIGAND_START",
            "LIGANDFILE",
            "LIGANDFILES",
            "LIGFORMAT",
            "LIGPREP",
            "LIGPREP_ARGS",
            "MACROCYCLE",
            "MACROCYCLE_OPTIONS",
            "MAX_ITERATIONS",
            "MAXATOMS",
            "MAXKEEP",
            "MAXREF",
            "MAXROTBONDS",
            "METAL_CUTOFF",
            "NENHANCED_SAMPLING",
            "NMAXRMSSYM",
            "NOSORT",
            "NREPORT",
            "NREQUIRED_CONS",
            "OUTPUTDIR",
            "PAIRDISTANCES",
            "PEPTIDE",
            "PHASE_DB",
            "PHASE_NCONFS",
            "PHASE_SUBSET",
            "POSE_DISPLACEMENT",
            "POSE_HTORSION",
            "POSE_OUTTYPE",
            "POSE_RMSD",
            "POSES_PER_LIG",
            "POSTDOCK",
            "POSTDOCK_ITMAX",
            "POSTDOCK_NPOSE",
            "POSTDOCK_SCITMAX",
            "POSTDOCK_XP_DELE",
            "POSTDOCKCG",
            "POSTDOCKLIGMIN",
            "POSTDOCKSTRAIN",
            "PRECISION",
            "PREMIN",
            "PREMINCG",
            "PREMINELEC",
            "PREMINITMAX",
            "RADIUS_RES_INTERACTION",
            "REF_LIGAND_FILE",
            "REFINDEX",
            "REPORT_CPU_TIME",
            "REWARD_INTRA_HBONDS",
            "RINGCONFCUT",
            "RINGONFLY",
            "SAMPLE_N_INVERSIONS",
            "SAMPLE_RINGS",
            "SCORE_INPUT_POSE",
            "SCORE_MINIMIZED_INPUT_POSE",
            "SCORING_CUTOFF",
            "SHAPE_ATOMS",
            "SHAPE_RESTRAIN",
            "SHAPE_TYPING",
            "SKIP_EPIK_METAL_ONLY",
            "STRAIN_GSFACTOR",
            "STRAIN_GSTHRESH",
            "STRAINELEC",
            "SUBSTRATE_PENAL_FILE",
            "USE_CONS",
            "USE_REF_LIGAND",
            "USECOMPMAE",
            "WRITE_CSV",
            "WRITE_RES_INTERACTION",
            "WRITE_TIMINGS_CSV",
            "WRITE_XP_DESC",
            "WRITEREPT",
        }

        result = {}
        current_block = None
        for linenum, line in enumerate(lines):
            if any(line.startswith(starter) for starter in block_starters):
                # Block start.
                current_block = line.strip()
                result[current_block] = {}
            elif line.strip() == "":
                # Empty line: close current block if any is open, and skip the line.
                current_block = None
            elif line.startswith(indent4):
                # Indented line inside the block.
                if current_block is None:
                    raise ValueError(
                        f"Unexpected indent outside of block for line {linenum}: {line}"
                    )
                kw, value = line.strip().split(sep=separator3, maxsplit=1)
                result[current_block][kw] = value.strip('"')
            elif any(line.startswith(kw) for kw in allowed_keywords):
                # Ordinary keywords.
                kw, value = line.strip().split(sep=separator3, maxsplit=1)
                result[kw] = value
            else:
                raise ValueError(
                    f"Unexpected line {linenum} in maestro input file: {line}"
                )

        return result



import os
from tempfile import mkdtemp

from pydantic import BaseModel
from rdkit import Chem
from copy import deepcopy
from typing import List, Tuple
from icolos.utils.execute_external.xtb import XTBExecutor

from icolos.utils.general.molecules import get_charge_for_molecule

from icolos.core.containers.compound import Conformer

from icolos.utils.enums.program_parameters import XTBEnum, XTBOutputEnum
from icolos.core.workflow_steps.step import _LE, _CTE
from icolos.core.workflow_steps.confgen.base import StepConfgenBase
from icolos.utils.general.parallelization import SubtaskContainer, Parallelizer

_EE = XTBEnum()
_COE = XTBOutputEnum()


class StepXTB(StepConfgenBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=XTBExecutor)
        self._check_backend_availability()

    def _parse_XTB_result(self, tmp_dirs: List, conformers: List[Conformer]):
        # XTB will output a variety of files to "dir_path"
        results = []
        for dir_path, conformer in zip(tmp_dirs, conformers):
            optimized_conformer_sdf = os.path.join(dir_path, _COE.XTBOPT_SDF)
            enum = conformer.get_enumeration_object()

            # as the energies are added as a tag, but we will use ours
            # note, that XTB is called to operate on one conformer at a time (which we will return here)
            mol_supplier = Chem.SDMolSupplier(optimized_conformer_sdf, removeHs=False)
            try:
                for mol in mol_supplier:
                    mol.SetProp(
                        _CTE.CONFORMER_ENERGY_TAG, mol.GetProp(_COE.TOTAL_ENERGY_TAG)
                    )
                    mol.ClearProp(_COE.TOTAL_ENERGY_TAG)
                    mol.SetProp(
                        _CTE.FORMAL_CHARGE_TAG, str(get_charge_for_molecule(mol))
                    )
                    enum.add_conformer(Conformer(conformer=mol), auto_update=True)
                results.append(_COE.SUCCESS)
            except:
                self._logger.log(
                    f"Failed to parse XTB results for conformer {conformer.get_index_string()}.",
                    _LE.WARNING,
                )
                results.append(_COE.FAILURE)
        return results

    def _prepare_batch(self, batch) -> Tuple:
        # first position is the input (SDF) file; the internal input at this stage is a list of molecules
        # -> write it to a temporary SDF file (undocumented input functionality) and add the path

        tmp_dirs = []
        input_files = []
        charges = []
        conformers = []
        for next_subtask_list in batch:
            tmp_dir = mkdtemp()
            tmp_dirs.append(tmp_dir)
            for (
                subtask
            ) in (
                next_subtask_list
            ):  # enforced as one since xtb can't handle multiple files in one call
                conformer = subtask.data
                conformers.append(conformer)
                input_file = self._prepare_temp_input(tmp_dir, conformer.get_molecule())
                charge = get_charge_for_molecule(conformer.get_molecule())

                charges.append(charge)
                input_files.append(input_file)
        return tmp_dirs, input_files, charges, conformers

    def _prepare_arguments(self, settings: List) -> List:

        # add flags
        for flag in self.settings.arguments.flags:
            settings.append(flag)

        # add parameters
        parameters = deepcopy(self.settings.arguments.parameters)

        # flatten the dictionary into a list for command-line execution
        for key in parameters.keys():
            settings.append(key)
            settings.append(parameters[key])
        return settings

    def _run_subjob(self, tmp_dir: str, input_file: str, charge: int) -> None:

        work_dir = os.getcwd()
        os.chdir(tmp_dir)

        arguments = [input_file, _EE.XTB_P, charge]
        arguments = self._prepare_arguments(
            arguments
        )  # add additional parameters from config

        self._backend_executor.execute(
            command=_EE.XTB, arguments=arguments, check=False
        )
        # for line in result.stdout.split("\n"):
        #     self._logger_blank.log(line, _LE.DEBUG)
        #     # print(line)
        os.chdir(work_dir)

    def _execute_xtb(self):
        xtb_parallelizer = Parallelizer(func=self._run_subjob)
        n = 1

        tmp_dirs = None
        while self._subtask_container.done() is False:

            next_batch = self._get_sublists(get_first_n_lists=self._get_number_cores())
            tmp_dirs, input_files, charges, conformers = self._prepare_batch(next_batch)

            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]

            self._logger.log(f"Executing xtb for batch {n}.", _LE.DEBUG)

            xtb_parallelizer.execute_parallel(
                tmp_dir=tmp_dirs,
                input_file=input_files,
                charge=charges,
            )

            results = self._parse_XTB_result(tmp_dirs, conformers)
            for sublist, result in zip(next_batch, results):
                assert len(sublist) == 1
                # TODO: this only works if max length sublist == 1, fine for now as that is all turbomole can handle
                for task in sublist:
                    if result == _COE.SUCCESS:
                        task.set_status_success()
                    else:
                        task.set_status_failed()

            n += 1
        self._remove_temporary(tmp_dirs)

    def execute(self):
        all_conformers = []
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if enumeration.get_conformers():
                    for conformer in enumeration.get_conformers():
                        all_conformers.append(conformer)
                enumeration.clear_conformers()
        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(all_conformers)
        self._execute_xtb()
        self._logger.log(
            f"Completed execution of XTB for {len(all_conformers)} conformers.",
            _LE.INFO,
        )



from icolos.loggers.steplogger import StepLogger
from icolos.utils.execute_external.sdconvert import SDConvertExecutor

from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.general.icolos_exceptions import StepFailed

_SEE = SchrodingerExecutablesEnum()
_LE = LoggingConfigEnum()


class SDConvertUtil:
    def __init__(self, prefix_execution: str = None, binary_location: str = None):
        self._logger = StepLogger()

        # initialize and check executor
        self.executor = SDConvertExecutor(
            prefix_execution=prefix_execution, binary_location=binary_location
        )
        if not self.executor.is_available():
            raise StepFailed("Cannot initialize sdconvert backend - abort.")
        self._logger.log(f"Checked sdconvert availability - valid.", _LE.DEBUG)

    def execute(self, arguments: list):
        execution_result = self.executor.execute(
            command=_SEE.SDCONVERT, arguments=arguments, check=True
        )
        if execution_result.returncode != 0:
            self._logger.log(
                f"Could not execute sdconvert (returncode != 0) with error: {execution_result.stderr}.",
                _LE.ERROR,
            )

    def mae2sdf(self, mae_file: str, sdf_file: str):
        arguments = [
            "".join([_SEE.SDCONVERT_I, _SEE.SDCONVERT_FORMAT_MAE]),
            mae_file,
            "".join([_SEE.SDCONVERT_O, _SEE.SDCONVERT_FORMAT_SD]),
            sdf_file,
        ]
        self.execute(arguments=arguments)

    def sdf2mae(self, sdf_file: str, mae_file: str):
        arguments = [
            "".join([_SEE.SDCONVERT_I, _SEE.SDCONVERT_FORMAT_SD]),
            sdf_file,
            "".join([_SEE.SDCONVERT_O, _SEE.SDCONVERT_FORMAT_MAE]),
            mae_file,
        ]
        self.execute(arguments=arguments)

    def sdf2pdb(self, sdf_file: str, pdb_file: str):
        arguments = [
            "".join([_SEE.SDCONVERT_I, _SEE.SDCONVERT_FORMAT_SD]),
            pdb_file,
            "".join([_SEE.SDCONVERT_O, _SEE.SDCONVERT_FORMAT_PDB]),
            sdf_file,
        ]
        self.execute(arguments=arguments)



import os
import subprocess
from typing import Tuple, List

from pydantic import BaseModel, PrivateAttr
from rdkit import Chem

from icolos.core.workflow_steps.schrodinger.base import StepSchrodingerBase
from icolos.utils.execute_external.macromodel import MacromodelExecutor

from icolos.utils.general.molecules import get_charge_for_molecule

from icolos.core.containers.compound import Enumeration, Conformer

from icolos.utils.enums.program_parameters import (
    MacromodelEnum,
)
from icolos.utils.enums.step_enums import StepMacromodelEnum
from icolos.core.workflow_steps.step import _LE, _CTE
from icolos.core.step_utils.sdconvert_util import SDConvertUtil

_EE = MacromodelEnum()
_MMSE = StepMacromodelEnum()


class StepMacromodel(StepSchrodingerBase, BaseModel):
    class Config:
        underscore_attrs_are_private = True

    _sdconvert_util = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=MacromodelExecutor)
        self._check_backend_availability()

        # prepare sdconvert utility
        self._sdconvert_util = SDConvertUtil(
            prefix_execution=self.execution.prefix_execution,
            binary_location=self.execution.binary_location,
        )

        # extend parameters with the COM file default, if not present
        if _MMSE.COM_FILE not in self.settings.arguments.parameters.keys():
            self.settings.arguments.parameters[_MMSE.COM_FILE] = _MMSE.COM_FILE_DEFAULT

    def _execute_macromodel(self, com_file: str) -> subprocess.CompletedProcess:
        self._logger.log(
            f"Executing MacroModel backend for com_file {com_file}.", _LE.DEBUG
        )
        arguments = []
        for key in self.settings.arguments.parameters.keys():
            # TODO: disentangle "special behaviour" for this key - move the com_file specification to a separate block
            #       in the configuration
            if key != _MMSE.COM_FILE:
                arguments.append(key)
                arguments.append(str(self.settings.arguments.parameters[key]))
        for flag in self.settings.arguments.flags:
            arguments.append(str(flag))
        arguments.append(com_file)
        self._apply_token_guard()
        result = self._backend_executor.execute(
            command=_EE.MACROMODEL, arguments=arguments, check=True
        )
        return result

    def _set_formal_charge(self, parameters: dict, molecule: Chem.Mol) -> dict:
        charge = get_charge_for_molecule(molecule)
        parameters[_EE.XTB_CHRG] = charge
        self._logger.log(f"Set charge for molecule to {charge}.", _LE.DEBUG)
        return parameters

    def _prepare_file_paths(self, tmp_dir: str) -> Tuple[str, str, str]:
        # generate the paths to the temporary files
        mae_input = os.path.join(tmp_dir, _MMSE.MAE_INPUT)
        mae_output = os.path.join(tmp_dir, _MMSE.MAE_OUTPUT)
        sdf_output = os.path.join(tmp_dir, _MMSE.SDF_OUTPUT)

        return mae_input, mae_output, sdf_output

    def _prepare_settings_file(self, tmp_dir: str) -> str:
        path_settings_file = os.path.join(tmp_dir, _MMSE.COM_FILE_PATH)

        # join the input and output paths (at the beginning of the COM file) and the
        # settings from either the default or the configuration together
        complete_com = "\n".join(
            [
                os.path.join(tmp_dir, _MMSE.MAE_INPUT),
                os.path.join(tmp_dir, _MMSE.MAE_OUTPUT),
                self.settings.arguments.parameters[_MMSE.COM_FILE],
            ]
        )
        with open(path_settings_file, "w") as f:
            f.writelines(complete_com)
        return path_settings_file

    def _prepare_run_files(
        self, tmp_dir: str, enumeration: Enumeration
    ) -> Tuple[str, str, str, str, str]:
        # generate the file paths (NOT populated yet)
        mae_input, mae_output, sdf_output = self._prepare_file_paths(tmp_dir)

        # write the input SDF file and translate it into Schrodingers native MAE format
        sdf_input = self._prepare_temp_input(tmp_dir, enumeration.get_molecule())
        self._sdconvert_util.sdf2mae(sdf_input, mae_input)

        # write out the settings file
        com_file = self._prepare_settings_file(tmp_dir)

        return sdf_input, mae_input, mae_output, sdf_output, com_file

    def _parse_macromodel_result(
        self, sdf_output: str, enumeration: Enumeration
    ) -> List[Conformer]:
        charge = str(
            get_charge_for_molecule(enumeration.get_molecule(), add_as_tag=False)
        )
        mol_supplier = Chem.SDMolSupplier(sdf_output, removeHs=False)
        conformers = []
        for mol_id, mol in enumerate(mol_supplier):
            # note, that formal charge information would be kept if available before (i.e. it retains tags)
            mol.SetProp(_CTE.FORMAL_CHARGE_TAG, charge)
            conformers.append(Conformer(conformer=mol))
        return conformers

    def execute(self):
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if not self._input_object_valid(enumeration):
                    continue

                # set up
                tmp_dir = self._move_to_temp_dir()

                # get the paths to the MAE and SDF input and output files and the COM file (settings)
                (
                    sdf_input,
                    mae_input,
                    mae_output,
                    sdf_output,
                    com_file,
                ) = self._prepare_run_files(tmp_dir=tmp_dir, enumeration=enumeration)

                # execute MacroModel, obtain the output SDF and switch back the working directory to what it was before
                _ = self._execute_macromodel(com_file=com_file)
                os.listdir(tmp_dir)
                self._sdconvert_util.mae2sdf(mae_file=mae_output, sdf_file=sdf_output)
                self._restore_working_dir()

                # parse output
                conformers = self._parse_macromodel_result(sdf_output, enumeration)
                enumeration.clear_conformers()
                enumeration.add_conformers(conformers=conformers, auto_update=True)
                self._logger.log(
                    f"Executed MacroModel and obtained {len(conformers)} conformers for enumeration {enumeration.get_index_string()}.",
                    _LE.INFO,
                )

                self._remove_temporary(tmp_dir)



import numpy as np
import os
from turtle import st
from typing import AnyStr, Dict, List
from pydantic import BaseModel
from icolos.core.containers.generic import GenericData
from icolos.utils.enums.program_parameters import GromacsEnum

from icolos.utils.enums.step_enums import StepGromacsEnum

_SGE = StepGromacsEnum()
_GE = GromacsEnum()


class AtomType(BaseModel):
    number: int
    a_type: str
    resi: int
    res: str
    atom: str
    cgnr: int
    charge: float
    mass: float


class GromacsState(BaseModel):
    class Config:
        arbitrary_types_allowed: True

    top_lines: List = []
    itps: Dict = {}
    posre: Dict = {}
    atomtypes: List = []
    chains: List = None
    forcefield: str = "amber03"
    water: str = "tip3p"
    system: List = []
    molecules: Dict = {}
    # hold the files as generic data objects, keyed by the index
    structures: Dict = {}
    tprs: Dict = {}
    trajectories: Dict = {}
    log: Dict = {}
    edr: Dict = {}
    cpt: Dict = {}
    ndx: List = []
    # store computed properties on the topology
    # {property: [val1, val2, val3]}
    properties: Dict[str, List] = {}

    def __init__(self, **data) -> None:
        super().__init__(**data)

    def parse(self, path: str, file: str = _SGE.STD_TOPOL):
        """
        Populate the fields from parsing a topol file (normally from gmx pdb2gmx)
        If a moleculetype has been defined in a single topology, this is separated into its own itp file
        """
        # get the system name
        with open(os.path.join(path, file), "r") as f:
            lines = f.readlines()

        # first check for included atomtypes to be extracted to their own itp files
        start_idx = None
        stop_idx = None
        for line in lines:
            if line.startswith(_GE.MOLECULETYPES):
                start_idx = lines.index(line)
            # go all the way to IFDEF POSRE
            elif line.startswith("#ifdef POSRES") and start_idx is not None:
                stop_idx = lines.index(line) + 3
                break
        if all([x is not None for x in (start_idx, stop_idx)]):
            itp_extract = lines[start_idx:stop_idx]
            itp_key = itp_extract[2].split()[0] + ".itp"
            self.itps[itp_key] = itp_extract
            lines = lines[:start_idx] + lines[stop_idx:]

        # extract itp files (not including forcefield stuff)
        for line in lines:
            if (
                line.startswith("#include")
                and ".itp" in line
                and all([item not in line for item in (".ff", "posre")])
            ):
                itp_file = line.split()[-1].replace('"', "")
                with open(os.path.join(path, itp_file), "r") as f:
                    itp_lines = f.readlines()
                self.itps[itp_file] = itp_lines
        # extract water model
        for line in lines:
            if line.startswith("#include") and ".ff/tip" in line:
                self.water = line.split("/")[-1].split(".")[0].replace('"', "")
                self.forcefield = line.split()[-1].split(".")[0].replace('"', "")
        start = lines.index([l for l in lines if l.startswith(_GE.SYSTEM)][0])
        stop = lines.index([l for l in lines[start:] if l.startswith(_GE.MOLECULES)][0])
        if not self.system:
            for line in lines[start + 1 : stop]:
                if not line.startswith(";") and line.strip():
                    self.system.append(line.strip())
        # excract molecules and counts
        for line in lines[stop + 1 :]:
            if not line.startswith(";") and line.strip():
                parts = line.split()
                self.molecules[parts[0].strip()] = parts[1].strip()

        self.top_lines = self.generate_base_topol_file()

    def write_topol(self, base_dir: str, file: str = "topol.top"):
        """
        Write a gromacs topology file, including its dependent itp files, to a dir
        """
        # update the base topol file
        self.generate_base_topol_file()
        with open(os.path.join(base_dir, file), "w") as f:
            for l in self.top_lines:
                f.write(l)
        for itp, lines in self.itps.items():
            if os.path.isfile(os.path.join(base_dir, itp)):
                os.remove(os.path.join(base_dir, itp))
            with open(os.path.join(base_dir, itp), "w") as f:
                for l in lines:
                    f.write(l)
        for itp, lines in self.posre.items():
            if os.path.isfile(os.path.join(base_dir, itp)):
                os.remove(os.path.join(base_dir, itp))
            with open(os.path.join(base_dir, itp), "w") as f:
                for l in lines:
                    f.write(l)

    def generate_base_topol_file(self):
        """
        Generates the main topology file
        """
        top_lines = []
        top_lines.append(f'#include "{self.forcefield}.ff/forcefield.itp"\n')
        # find any new atomtypes in the parametrised components, slot these in first
        if not self.atomtypes:
            self.atomtypes = self.collect_atomtypes()
        top_lines += self.atomtypes

        # now include the itp files
        for file in self.itps.keys():
            top_lines.append(f'#include "{file}"\n')
            # pdb2gmx appends posre control info to the bottom of any itp files
            if all([x not in file for x in ("Protein", "DNA", "RNA")]):
                # these are handled independently in the itp files
                top_lines.append(f'#ifdef POSRES\n#include "posre_{file}"\n#endif\n')

        # add water model
        top_lines.append(f'#include "{self.forcefield}.ff/{self.water}.itp"\n')
        top_lines.append(_SGE.WATER_POSRE)

        # add ions itp
        top_lines.append(f'#include "{self.forcefield}.ff/ions.itp"\n')

        top_lines.extend(self._construct_block(_GE.SYSTEM, self.system))
        top_lines.extend(self._construct_block(_GE.MOLECULES, self.molecules))
        self.top_lines = top_lines

    def add_itp(self, path, files: List[str], gen_posre: bool = True) -> None:
        for file in files:
            with open(os.path.join(path, file), "r") as f:
                lines = f.readlines()
            self.itps[file] = lines
            if gen_posre:
                # also generate a posre file
                self.generate_posre(path, file)

        self.generate_base_topol_file()

    def add_molecule(self, name: str, num: int = 1):
        self.molecules[name] = num
        self.generate_base_topol_file()

    def add_posre(self, path, files: List[str]) -> None:
        for file in files:
            with open(os.path.join(path, file), "r") as f:
                lines = f.readlines()

            self.posre[file] = lines
        self.generate_base_topol_file()

    def generate_posre(self, path: str, itp_file: str, force: int = 1000):
        stub = itp_file.split(".")[0]
        with open(os.path.join(path, itp_file), "r") as f:
            lines = f.readlines()
        start_idx = None
        stop_idx = None
        for line in lines:
            if line.startswith(_GE.ATOMS_DIRECTIVE):
                start_idx = lines.index(line) + 1
            elif line.startswith(_GE.BONDS) and start_idx is not None:
                stop_idx = lines.index(line)
                break
        lines = [
            l for l in lines[start_idx:stop_idx] if not l.startswith(";") and l.strip()
        ]
        args = ["number", "a_type", "resi", "res", "atom", "cgnr", "charge", "mass"]
        atoms = []
        for line in lines:
            args_dict = {}
            for arg, val in zip(args, line.split(";")[0].split()):
                args_dict[arg] = val
            atoms.append(AtomType(**args_dict))

        posre = "posre_" + stub + ".itp"
        out_path = os.path.join(path, posre)
        header = "\n[ position_restraints ]\n; atom  type      fx      fy      fz\n"
        written_lines = [header]
        with open(out_path, "w") as f:
            f.write(header)
            for atom in atoms:
                if not atom.a_type.upper().startswith("H"):
                    line = (
                        f"{atom.number:>6d}     1 {force:>5d} {force:>5d} {force:>5d}\n"
                    )
                    f.write(line)
                    written_lines.append(line)
        self.posre[posre] = written_lines

    def _construct_block(self, header, items) -> List:
        if not header.endswith("\n"):
            header += "\n"
        block = [header]
        if isinstance(items, dict):
            for key, value in items.items():
                block.append(f"{key}   {value}\n")
        else:
            for i in items:
                block.append(i + "\n")
        return block

    def collect_atomtypes(self) -> List:
        """
        Iterate over the itp files, strip any newly defined atomtypes, append to their own atomtypes.itp file, include this just after the forcefield include
        """
        atomtype_lines = []
        # read the topol file, identify all the itp files it is #including
        for itp, file_lines in self.itps.items():
            selection, remaining = self._extract_atomtype(file_lines)
            atomtype_lines.extend(selection)
            self.itps[itp] = remaining
        if atomtype_lines:
            atomtype_lines = self._remove_duplicate_atomtypes(atomtype_lines)
        return atomtype_lines

    def _extract_atomtype(self, lines: List) -> List[AnyStr]:
        """
        Pull the atomtype lines out of the topol file and return them as a list, write the sanitised itp file to directory
        """

        start_index = None
        stop_index = None
        selection = []
        remaining = []
        for idx, line in enumerate(lines):
            if _GE.ATOMTYPES in line:
                start_index = idx
            if _GE.MOLECULETYPES in line:
                stop_index = idx
        if start_index is not None:
            selection = lines[start_index:stop_index]
            remaining = lines[:start_index]

        remaining.extend(lines[stop_index:])
        return selection, remaining

    def _remove_duplicate_atomtypes(self, atomtypes: List):
        output = [atomtypes[0]]
        for line in atomtypes:
            if line not in output:
                output.append(line)
        return output

    def set_topol(self, path: str, file: str = _SGE.STD_TOPOL):
        """
        When solvate or genion produce a modified topol, read this into the topology lines
        """
        with open(os.path.join(path, file), "r") as f:
            self.top_lines = f.readlines()

    def set_cpt(self, path: str, file: str = _SGE.STD_CPT, index: int = 0):
        with open(os.path.join(path, file), "rb") as f:
            lines = f.read()

        cpt_file = GenericData(file_name=file, file_data=lines)
        self.cpt[index] = cpt_file

    def set_edr(self, path: str, file: str = _SGE.STD_EDR, index: int = 0):
        with open(os.path.join(path, file), "rb") as f:
            lines = f.read()

        edr_file = GenericData(file_name=file, file_data=lines)
        self.edr[index] = edr_file

    def set_structure(
        self, path: str, file: str = _SGE.STD_STRUCTURE, sanitize=False, index: int = 0
    ):
        with open(os.path.join(path, file), "r") as f:
            lines = f.readlines()
        if sanitize:
            lines = [l for l in lines if any([l.startswith(idx) for idx in _GE.ATOMS])]

        struct = GenericData(file_name=file, file_data=lines)
        self.structures[index] = struct

    def set_log(self, path: str, file: str = _SGE.STD_LOG, index: int = 0):
        with open(os.path.join(path, file), "r") as f:
            lines = f.readlines()
        log = GenericData(file_name=file, file_data=lines)
        self.log[index] = log

    def set_tpr(self, path: str, file: str = _SGE.STD_TPR, index: int = 0):
        with open(os.path.join(path, file), "rb") as f:
            data = f.read()
        data = GenericData(file_name=file, file_data=data, file_id=index)
        self.tprs[index] = data

    def write_tpr(self, path: str, file: str = _SGE.STD_TPR, index: int = 0):
        tpr = self.tprs[index]
        tpr.write(os.path.join(path, file), join=False)

    def write_log(self, path: str, file: str = _SGE.STD_LOG, index: int = 0):
        log = self.log[index]
        log.write(os.path.join(path, file), join=False)

    def write_cpt(self, path: str, file: str = _SGE.STD_CPT, index: int = 0):
        cpt = self.cpt[index]
        cpt.write(os.path.join(path, file), join=False)
    
    def write_edr(self, path: str, file: str = _SGE.STD_EDR, index: int = 0):
        edr = self.edr[index]
        edr.write(os.path.join(path, file), join=False)

    def set_ndx(self, path: str, file: str = _SGE.STD_INDEX):
        with open(os.path.join(path, file), "r") as f:
            self.ndx = f.readlines()

    def write_ndx(self, path: str, file: str = _SGE.STD_INDEX):
        with open(os.path.join(path, file), "w") as f:
            f.writelines(self.ndx)

    def write_structure(
        self, path: str, file: str = _SGE.STD_STRUCTURE, index: int = 0
    ):
        structure = self.structures[index]
        path = os.path.join(path, file)
        structure.write(path, join=False)

    def set_trajectory(self, path: str, file: str = _SGE.STD_XTC, index: int = 0):
        # depending on mdp settings, some runs will not produce an xtc file, only trr
        if os.path.isfile(os.path.join(path, file)):
            with open(os.path.join(path, file), "rb") as f:
                data = f.read()
            data = GenericData(file_name=file, file_data=data, file_id=index)
            self.trajectories[index] = data

    def write_trajectory(self, path: str, file: str = _SGE.STD_XTC, index: int = 0):
        traj = self.trajectories[index]

        path = os.path.join(path, file)
        traj.write(path, join=False)

    def append_structure(
        self, path: str, file: str, sanitize=False, index: int = 0
    ) -> None:
        with open(os.path.join(path, file), "r") as f:
            lines = f.readlines()
        if sanitize:
            lines = [l for l in lines if any([l.startswith(idx) for idx in _GE.ATOMS])]
        data = self.structures[index].get_data() + lines
        self.structures[index].set_data(data)

    def write_props(self, path: str):
        """
        writes summary of properties for all trajectories to a file
        """
        prop_string = [f"{key}: {value}" for key, value in self.properties.items()]
        avg_prop_string = [
            f"{key}: {np.mean(value)}" for key, value in self.properties.items()
        ]

        output_string = f"""
ICOLOS PROPERTY SUMMARY
### Computed properties properties: {self.properties.keys()}

### Properties
{prop_string}


### Average Properties
{avg_prop_string} 
        """

        with open(os.path.join(path, "ICOLOS_PROPS.dat"), "w") as f:
            f.write(output_string)

    def __str__(self) -> str:
        return f"Gromacs Topology object: System: {self.system} | Molecules: {[m for m in self.molecules]} | FF: {self.forcefield} | itp files: {[f for f in self.itps.keys()]} | posre files {[f for f in self.posre.keys()]}"

    def __repr__(self) -> str:
        return self.__str__()



class WriteOutEnum:

    RDKIT_NAME = "_Name"
    INDEX_STRING = "index_string"
    COMPOUND_NAME = "compound_name"

    # REINVENT-compatible JSON write-out
    JSON_RESULTS = "results"
    JSON_NAMES = "names"
    JSON_NA = ""
    JSON_VALUES = "values"
    JSON_VALUES_KEY = "values_key"

    SDF = "sdf"
    PDB = "pdb"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class RunVariablesEnum:

    PREFIX = "["
    POSTFIX = "]"
    COMPOUND_ID = "compound_id"
    ENUMERATION_ID = "enumeration_id"
    CONFORMER_ID = "conformer_id"
    COMPOUND_NAME = "compound_name"
    ENUMERATION_STRING = "enumeration_string"
    CONFORMER_STRING = "conformer_string"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")



from copy import deepcopy

from pydantic import BaseModel
from rdkit import Chem, RDLogger
from rdkit.Chem import AllChem

from icolos.core.containers.compound import Conformer
from icolos.utils.general.icolos_exceptions import StepFailed
from icolos.utils.enums.step_enums import StepEmbeddingEnum
from icolos.core.workflow_steps.io.base import StepIOBase

from icolos.core.workflow_steps.step import _LE
from icolos.utils.general.convenience_functions import *
from icolos.utils.smiles import to_mol

_SEE = StepEmbeddingEnum()


class StepEmbedding(StepIOBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # extend parameters with defaults
        if _SEE.EMBED_AS not in self.settings.additional.keys():
            self.settings.additional[_SEE.EMBED_AS] = _SEE.EMBED_AS_ENUMERATIONS
            self._logger.log(
                f'No embedding level specified, defaulting to "{_SEE.EMBED_AS_ENUMERATIONS}".',
                _LE.INFO,
            )

    def _smile_to_molecule(self, smile: str) -> Chem.Mol:
        mol = to_mol(smile)
        if mol is None:
            self._logger.log(
                f"The smile {smile} could not be transformed into a molecule and will be skipped.",
                _LE.WARNING,
            )
        return mol

    def _embed_with_RDKit(self, smile: str, parameters: dict) -> Chem.Mol:
        molecule = self._smile_to_molecule(smile)

        # deactivate logger to suppress "missing Hs messages"
        RDLogger.DisableLog("rdApp.*")
        try:
            embed_code = AllChem.EmbedMolecule(
                molecule, randomSeed=42, useRandomCoords=True
            )
        # TODO: what's the exception here?
        except:
            self._logger.log(
                f'Could not embed molecule with SMILES "{smile}", critical error in "RDkit".',
                _LE.WARNING,
            )
            return None

        status = 0
        if embed_code != -1:
            status = AllChem.UFFOptimizeMolecule(molecule, maxIters=600)
            if status == 1:
                self._logger.log(
                    f"The 3D coordinate generation of molecule {smile} did not converge in time.",
                    _LE.WARNING,
                )
        else:
            self._logger.log(
                f"Could not embed molecule {smile} - no 3D coordinates have been generated.",
                _LE.WARNING,
            )
        RDLogger.EnableLog("rdApp.*")

        # add hydrogens to the molecule (if specified)
        if nested_get(parameters, [_SEE.RDKIT_PROTONATE], default=True):
            molecule = Chem.AddHs(molecule, addCoords=True)

        if embed_code != -1 and status == 0:
            return molecule
        else:
            return None

    def _get_embedding_method(self, parameters: dict) -> str:
        method = nested_get(parameters, [_SEE.METHOD], default=None)
        if method is None:
            error = "Embedding method not set."
            self._logger.log(error, _LE.ERROR)
            raise StepFailed(error)
        return method.upper()

    def _embed_molecule(self, smile: str, parameters: dict) -> Chem.Mol:
        method = self._get_embedding_method(parameters)
        if method == _SEE.METHOD_RDKIT:
            return self._embed_with_RDKit(smile, parameters)
        else:
            self._logger.log(
                f"Specified embedding method {method} not available.", _LE.ERROR
            )

    def execute(self):
        # TODO: REFACTOR
        parameters = deepcopy(self.settings.arguments.parameters)
        embed_as = self.settings.additional[_SEE.EMBED_AS]
        for compound in self.get_compounds():
            if embed_as == _SEE.EMBED_AS_ENUMERATIONS:
                enum_buffer = deepcopy(compound.get_enumerations())
                compound.clear_enumerations()
                for enumeration in enum_buffer:
                    enumeration.clear_molecule()
                    enumeration.clear_conformers()
                    molecule = self._embed_molecule(
                        smile=enumeration.get_smile(), parameters=parameters
                    )
                    if molecule is not None:
                        enumeration.set_molecule(molecule)
                        compound.add_enumeration(enumeration)
                self._logger.log(
                    f"Embedding for compound {compound.get_index_string()} (name: {compound.get_name()}) completed ({len(compound)} of {len(enum_buffer)} enumerations successful).",
                    _LE.INFO,
                )
            elif embed_as == _SEE.EMBED_AS_CONFORMERS:
                # TODO: double-check this bit
                for enumeration in compound.get_enumerations():
                    enumeration.clear_conformers()
                    molecule = self._embed_molecule(
                        smile=enumeration.get_smile(), parameters=parameters
                    )
                    if molecule is not None:
                        conformer = Conformer(
                            conformer=molecule, enumeration_object=enumeration
                        )
                        enumeration.add_conformer(conformer, auto_update=True)
                number_successful = len(
                    [
                        True
                        for enum in compound.get_enumerations()
                        if enum[0].get_molecule() is not None
                    ]
                )
                self._logger.log(
                    f"Embedding for compound {compound.get_index_string()} (name: {compound.get_name()}) completed ({number_successful} of {len(compound)} enumerations successful).",
                    _LE.INFO,
                )
            else:
                ValueError(
                    f'Value "{embed_as}" for parameter "embed_as" not supported.'
                )



import time
from typing import Dict

from pydantic import BaseModel, PrivateAttr

from icolos.utils.execute_external.execute import Executor

from icolos.loggers.steplogger import StepLogger

from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum

_EE = SchrodingerExecutablesEnum()
_LE = LoggingConfigEnum()


class TokenGuardParameters(BaseModel):
    prefix_execution: str = None
    binary_location: str = None
    token_pools: Dict
    wait_interval_seconds: int = 30
    wait_limit_seconds: int = 0


class SchrodingerLicenseTokenGuard(BaseModel):
    """Class that checks, whether enough tokens to execute Schrodinger binaries are available."""

    token_guard: TokenGuardParameters

    class Config:
        underscore_attrs_are_private = True

    _logger = PrivateAttr()
    _executor = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)

        self._logger = StepLogger()

        # initialize the executor for all "Schrodinger" related calls and also check if it is available
        self._executor = Executor(
            prefix_execution=self.token_guard.prefix_execution,
            binary_location=self.token_guard.binary_location,
        )

    def _get_token_pool_info(self, licadmin_output: list, token_pool: str) -> dict:
        result = {"found": False}
        for line in licadmin_output:
            if token_pool in line:
                parts = line.split(" ")
                if len(parts) == 16:
                    result["total"] = int(parts[6])
                    result["available"] = int(parts[6]) - int(parts[12])
                    result["found"] = True
                break
        return result

    def _check_licstat_output(self, licadmin_output: list) -> bool:
        all_pools_available = True
        for pool_key, pool_token_numbers in self.token_guard.token_pools.items():
            pool_status = self._get_token_pool_info(licadmin_output, pool_key)
            if pool_status["found"]:
                if pool_status["available"] >= pool_token_numbers:
                    self._logger.log(
                        f"Enough tokens available ({pool_status['available']}) to satisfy requirement ({pool_token_numbers} free tokens) for pool {pool_key}.",
                        _LE.DEBUG,
                    )
                else:
                    self._logger.log(
                        f"Not enough tokens available ({pool_status['available']}) to satisfy requirement ({pool_token_numbers} free tokens) for pool {pool_key}.",
                        _LE.DEBUG,
                    )
                    all_pools_available = False
            else:
                all_pools_available = False
                self._logger.log(
                    f"Could not find information on token pool {pool_key}.", _LE.WARNING
                )
        return all_pools_available

    def _get_licstat_output(self):
        result = self._executor.execute(
            command=_EE.LICADMIN, arguments=[_EE.LICADMIN_STAT], check=True
        )
        if result.returncode != 0:
            self._logger.log(
                f"Could not execute the Schrodinger license token guard - do you need to export the licadmin path?",
                _LE.WARNING,
            )
        return result.stdout.split("\n")

    def guard(self) -> bool:
        # loop over the token pools until they are all satisfied or the time limit has run out
        counter = 0
        success = False
        while True:
            if (
                self.token_guard.wait_limit_seconds != 0
                and (counter * self.token_guard.wait_interval_seconds)
                >= self.token_guard.wait_limit_seconds
            ):
                self._logger.log(
                    f"Wait period ({self.token_guard.wait_limit_seconds} seconds) set for Schrodinger token guard has been exceeded.",
                    _LE.ERROR,
                )
                break

            # reload the output from "licadmin"
            # at this stage, the output from licadmin is a list of strings
            licadmin_output = self._get_licstat_output()

            all_pools_available = self._check_licstat_output(
                licadmin_output=licadmin_output
            )
            if all_pools_available:
                self._logger.log(
                    "All token pool requirements for Schrodinger have been met - proceeding.",
                    _LE.DEBUG,
                )
                success = True
                break
            else:
                time.sleep(self.token_guard.wait_interval_seconds)
                counter = counter + 1

        return success



from icolos.utils.enums.step_enums import StepBaseEnum, StepGromacsEnum
from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from icolos.utils.execute_external.gromacs import GromacsExecutor
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE
import os

_GE = GromacsEnum()
_SGE = StepGromacsEnum()
_SBE = StepBaseEnum


class StepGMXCluster(StepGromacsBase, BaseModel):
    """
    Execute gmx cluster on a trajectory
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)
        self._check_backend_availability()

    def execute(self):
        tmp_dir = self._make_tmpdir()
        topol = self.get_topol()
        self.write_input_files(tmp_dir, topol=topol)

        # give the option to run a make_ndx step preceding clustering to facilitate clustering on custom groups
        if _SGE.INDEX_FLAG in self.settings.arguments.parameters.keys():
            assert (
                _SGE.STD_INDEX in os.listdir(tmp_dir)
                or self.settings.additional[_SGE.MAKE_NDX_COMMAND] is not None
            )
            if _SGE.STD_INDEX not in os.listdir(tmp_dir):
                try:
                    ndx_arguments = [
                        "-f",
                        _SGE.STD_STRUCTURE,
                        "-o",
                        _SGE.STD_INDEX,
                    ]
                    self._backend_executor.execute(
                        command=_GE.MAKE_NDX,
                        arguments=ndx_arguments,
                        location=tmp_dir,
                        check=True,
                        pipe_input=self.construct_pipe_arguments(
                            tmp_dir, self.settings.additional[_SGE.MAKE_NDX_COMMAND]
                        ),
                    )

                except KeyError:
                    raise KeyError(
                        "If the index flag was specified, you must provide the ndx command in additional "
                        "settings"
                    )

        flag_dict = {
            "-s": _SGE.STD_TPR,
            "-f": _SGE.STD_XTC,
            "-cl": "clusters.pdb",
            "-clid": "cluster_id.xvg",
        }
        arguments = self._parse_arguments(flag_dict=flag_dict)

        result = self._backend_executor.execute(
            command=_GE.CLUSTER,
            arguments=arguments,
            location=tmp_dir,
            check=True,
            pipe_input=self.construct_pipe_arguments(
                tmp_dir, self.settings.additional[_SBE.PIPE_INPUT]
            ),
        )

        for line in result.stdout.split("\n"):
            self._logger_blank.log(line, _LE.DEBUG)
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



import os
import abc
import subprocess
from shlex import quote

from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.loggers.steplogger import StepLogger

_LE = LoggingConfigEnum()


class ExecutorBase(metaclass=abc.ABCMeta):
    """Virtual base class for the general and program-specific executors."""

    def __init__(self, prefix_execution=None, binary_location=None):
        # if something needs to be attached to the execution string each time, store it here; if not, value is "None"
        self._prefix_execution = prefix_execution
        self._binary_location = binary_location
        # initialise from the step with self.execution.resource dict

    @abc.abstractmethod
    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        # to avoid security issues, escape the arguments
        arguments = [quote(str(arg)) for arg in arguments]

        # allow for piped input to be passed to binaries
        if pipe_input is not None:
            # pipe_input = self._parse_pipe_input(pipe_input)
            command = pipe_input + " | " + command

        # check, if command (binary) is to be found at a specific location (rather than in $PATH)
        if self._binary_location is not None:
            command = os.path.join(self._binary_location, command)

        # check, if the something needs to be added before the execution of the "rDock" command
        if self._prefix_execution is not None:
            command = self._prefix_execution + " && " + command

        # execute; if "location" is set, change to this directory and execute there
        complete_command = command + " " + " ".join(str(e) for e in arguments)
        if "_asl" not in complete_command:
            complete_command = [complete_command.replace("'", "")]
        old_cwd = os.getcwd()
        if location is not None:
            os.chdir(location)
        # determine whether this is to be run using local resources or as a batch job
        result = subprocess.run(
            complete_command,
            check=False,  # use the manual check to provide better debugginf information than subprocess
            # convert output to string (instead of byte array)
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        # for line in result.stdout.split("\n"):
        #     print(line)
        if check:
            if result.returncode != 0:
                raise subprocess.SubprocessError(
                    f"Subprocess returned non-zero exit status:\n{complete_command}\nReturn code:\n{result.returncode}\nSTDERR:\n{result.stderr}\nSTDOUT:\n{result.stdout}"
                )
        os.chdir(old_cwd)
        return result

    @abc.abstractmethod
    def is_available(self):
        raise NotImplementedError("Overwrite this method in the child class.")


class Executor(ExecutorBase):
    """For execution of command-line programs that do not have any specific executor themselves."""

    def __init__(self, prefix_execution=None, binary_location=None):
        super().__init__(
            prefix_execution=prefix_execution,
            binary_location=binary_location,
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=location,
            pipe_input=pipe_input,
        )

    def is_available(self):
        raise NotImplementedError(
            "Cannot reliably check, whether a random program executes properly - do not use."
        )


def execution_successful(output: str, success_str: str) -> bool:
    return True if success_str in output else False



from icolos.utils.execute_external.execute import ExecutorBase
from icolos.utils.enums.program_parameters import InducedFitEnum

_IFE = InducedFitEnum()


class IFDExecutor(ExecutorBase):
    def __init__(self, prefix_execution=None, binary_location=None) -> None:
        super().__init__(
            prefix_execution=prefix_execution, binary_location=binary_location
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):

        if command not in [_IFE.IFD_EXEC]:
            raise AssertionError(
                "Commmand must be recognised in the internal dictionary"
            )

        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=location,
            pipe_input=pipe_input,
        )

    def is_available(self):
        try:
            result = self.execute(
                command=_IFE.IFD_EXEC, arguments=[_IFE.IFD_HELP], check=True
            )

            if _IFE.IFD_HELP_ID in result.stdout:
                return True
            return False
        except Exception as e:
            return False



from setuptools import setup, find_namespace_packages
from pathlib import Path

long_description = (Path(__file__).parent / "README.md").read_text()

setup(
    name="icolos",
    maintainer="Christian Margreitter, Harry Moore",
    version="1.10.0",
    url="https://github.com/MolecularAI/Icolos",
    packages=find_namespace_packages(where="src"),
    package_dir={"": "src"},
    description="Icolos Workflow Manager",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    entry_points={
        "console_scripts": [
            "icolos = icolos.scripts.executor:main",
            "validator = icolos.scripts.validator:main",
            "sdf2smi = icolos.scripts.sdf2smi:main",
        ]
    },
)



import numpy as np
from subprocess import CompletedProcess
import tempfile
from icolos.core.containers.generic import GenericData
from typing import AnyStr, List
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from pydantic import BaseModel
from icolos.utils.enums.step_enums import StepGromacsEnum
from icolos.core.workflow_steps.step import _LE
from icolos.utils.enums.program_parameters import GromacsEnum
import os
from icolos.utils.execute_external.gromacs import GromacsExecutor
from icolos.utils.general.files_paths import attach_root_path

_SGE = StepGromacsEnum()
_GE = GromacsEnum()


class StepGMXmmpbsa(StepGromacsBase, BaseModel):
    """
    Execute gmx_MMPBSA, calculates binding free energy of
    protein-ligand complex using single trajectory approximation,
    using Amber's mmpbsa.py script
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(GromacsExecutor)
        self._check_backend_availability()

    def _get_arg(self, ext) -> AnyStr:
        return self.data.generic.get_argument_by_extension(ext)

    def _generate_amber_input_file(self) -> None:
        input_file = (
            self.settings.additional[_SGE.INPUT_FILE]
            if _SGE.INPUT_FILE in self.settings.additional.keys()
            else None
        )
        # Normally the user should provide an input file to control the mmgbsa protocol
        if input_file is not None and os.path.isfile(input_file):
            self._logger.log(
                f"Using provided AMBER input file at {self.settings.additional[_SGE.INPUT_FILE]}",
                _LE.DEBUG,
            )
            with open(input_file, "r") as f:
                template = GenericData(file_name="mmpbsa.in", file_data=f.read())
        else:
            self._logger.log("No input file found, defaulting to template", _LE.WARNING)
            # parses user arguments and creates the formatted amber input file from the user specification
            with open(attach_root_path(_SGE.DEFAULT_MMPBSA_IN), "r") as f:
                template = GenericData(file_name="mmpbsa.in", file_data=f.read())

        self.data.generic.add_file(template)

    def _parse_arguments(self, flag_dict: dict) -> List:
        args = []
        for flag in self.settings.arguments.flags:
            if flag != "-O":
                args.append(flag)
        for key, value in self.settings.arguments.parameters.items():
            args.append(key)
            args.append(value)
        for key, value in flag_dict.items():
            if key not in args:
                args.append(key)
                args.append(value)

        # capture output
        return args

    def _run_mmpbsa(self, args, tmp_dir) -> CompletedProcess:
        command = _GE.MMPBSA
        threads = self._get_additional_setting(_SGE.THREADS, default=1)
        if threads > 1:
            command = f"mpirun -np {threads} " + command
        self._logger.log(
            f"Executing mmgbsa calculation with {threads} thread(s)", _LE.DEBUG
        )
        result = self._backend_executor.execute(
            command=command, arguments=args, check=True, location=tmp_dir
        )
        for line in result.stdout.split("\n"):
            self._logger_blank.log(line, _LE.INFO)
        for line in result.stderr.split("\n"):
            self._logger_blank.log(line, _LE.INFO)

        return result

    def _parse_coupling_groups(self, tmp_dir) -> AnyStr:
        # parse the coupling groups to their indexes
        output = []
        pipe_input = self.settings.additional[_SGE.COUPLING_GROUPS]

        arguments = ["-f", _SGE.STD_STRUCTURE]
        if [f for f in os.listdir(tmp_dir) if f.endswith("ndx")]:
            arguments.extend(["-n", "index.ndx"])
        else:
            arguments.extend(["-o", "index.ndx"])

        result = self._backend_executor.execute(
            command=_GE.MAKE_NDX,
            arguments=arguments,
            location=tmp_dir,
            check=True,
            pipe_input='echo -e "q"',
        )
        for param in pipe_input.split():
            for line in result.stdout.split("\n"):
                parts = line.split()
                if param in line and parts[1] == param:
                    output.append(parts[0])
                    break
        self._logger.log(f"Resolved coupling groups {output}", _LE.DEBUG)
        return " ".join(output)

    def _get_file_from_dir(self, tmp_dir: str, ext: str) -> AnyStr:
        file = [f for f in os.listdir(tmp_dir) if f.endswith(ext)]
        assert len(file) == 1
        return file[0]

    def parse_results(self, wkdirs: List[str]):
        """Parse the .dat files from the workdirs, extract results and attach to topol

        :param List[str] wkdirs: list of workdirs to walk through
        """
        for wkdir in wkdirs:
            results_file = os.path.join(wkdir, "FINAL_RESULTS_MMPBSA.dat")
            with open(results_file, "r") as f:
                lines = f.readlines()
            # for now just parse the final energy value
            result = float([l for l in lines if "DELTA TOTAL" in l][0].split()[2])
            try:
                self.get_topol().properties[_SGE.MMGBSA_DG].append(result)
            except KeyError:
                self.get_topol().properties[_SGE.MMGBSA_DG] = [result]
        avg_dG = np.mean(self.get_topol().properties[_SGE.MMGBSA_DG])
        self._logger.log(f"Average dG = {avg_dG}", _LE.INFO)
        # write the invididual lines to the tmpdir
        parent_dir = os.path.dirname(wkdirs[0])
        with open(os.path.join(parent_dir, "results_summary.dat"), "w") as f:
            for idx, val in enumerate(self.get_topol().properties[_SGE.MMGBSA_DG]):
                f.write(str(idx) + "," + str(val))

    def execute(self) -> None:
        tmp_dir = self._make_tmpdir()
        topol = self.get_topol()
        self._generate_amber_input_file()
        self.write_input_files(tmp_dir, topol=topol)

        # form any required coupling groups with make_ndx_command before parsing coupling groups
        # e.g. combine protein + cofactor
        ndx_commands = (
            self.settings.additional[_SGE.MAKE_NDX_COMMAND]
            if _SGE.MAKE_NDX_COMMAND in self.settings.additional.keys()
            else None
        )
        if ndx_commands is not None:
            # can run make_ndx multiple times for complex cases, each set of pipe imput must be separated by a semicolon
            for args in ndx_commands.split(";"):
                self._add_index_group(tmp_dir=tmp_dir, pipe_input=args)
        wkdirs = [
            tempfile.mkdtemp(dir=tmp_dir) for _ in range(len(topol.trajectories.keys()))
        ]

        for i, wkdir in enumerate(wkdirs):
            self._logger.log(
                f"Running MMGBSA on trajectory {i+1} (of {len(wkdirs)})", _LE.DEBUG
            )
            topol.write_structure(wkdir, index=i)
            topol.write_tpr(wkdir, index=i)
            topol.write_trajectory(wkdir, index=i)

            flag_dict = {
                "-i": os.path.join(tmp_dir, _SGE.MMPBSA_IN),
                "-cs": _SGE.STD_TPR,
                "-cg": self._parse_coupling_groups(tmp_dir),
                "-ci": os.path.join(tmp_dir, _SGE.STD_INDEX),
                "-ct": _SGE.STD_XTC,
                "-cp": os.path.join(tmp_dir, _SGE.STD_TOPOL),
                # do not attempt to open the results in the GUI afterwards
                "-nogui": "",
            }

            flag_list = self._parse_arguments(flag_dict=flag_dict)

            self._run_mmpbsa(args=flag_list, tmp_dir=wkdir)
        self.parse_results(wkdirs)

        # parse and delete generated output
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



import pandas as pd
from typing import List, Tuple

from pydantic import BaseModel

from icolos.core.containers.compound import Conformer

from icolos.utils.enums.step_enums import StepClusteringEnum
from icolos.core.workflow_steps.step import _LE
from icolos.core.workflow_steps.calculation.base import StepCalculationBase

from sklearn.cluster import KMeans

_SC = StepClusteringEnum()


class StepClustering(StepCalculationBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # extend parameters
        if _SC.N_CLUSTERS not in self.settings.arguments.parameters.keys():
            self.settings.arguments.parameters[_SC.N_CLUSTERS] = 3
        if _SC.MAX_ITER not in self.settings.arguments.parameters.keys():
            self.settings.arguments.parameters[_SC.MAX_ITER] = 300
        if _SC.TOP_N_PER_SOLVENT not in self.settings.additional.keys():
            self.settings.additional[_SC.TOP_N_PER_SOLVENT] = 3

    def _get_nclusters_and_top_n(self, len_conformers: int) -> Tuple[int, int]:
        n_clusters = self.settings.arguments.parameters[_SC.N_CLUSTERS]
        if n_clusters > len_conformers:
            n_clusters = len_conformers
            self._logger.log(
                f"Set number of clusters to {n_clusters} because not enough observations were provided.",
                _LE.DEBUG,
            )
        top_n_per_solvent = self.settings.additional[_SC.TOP_N_PER_SOLVENT]
        if top_n_per_solvent > len_conformers:
            top_n_per_solvent = len_conformers
            self._logger.log(
                f'Set number of "top_N_per_solvent" to {top_n_per_solvent} because not enough observations were provided.',
                _LE.DEBUG,
            )
        return n_clusters, top_n_per_solvent

    def _generate_feature_dataframe(self, conformers: List[Conformer]) -> pd.DataFrame:
        features = self.settings.additional[_SC.FEATURES]
        df_features = pd.DataFrame(columns=features)
        for conf in conformers:
            new_row = {}
            for feature in features:
                new_row[feature] = float(conf.get_molecule().GetProp(feature))
            df_features = df_features.append(new_row, ignore_index=True)
        return df_features

    def _get_representative_conformers(
        self, cluster_set: List[Tuple[int, Conformer]]
    ) -> List[int]:
        # for each selection (e.g. solvent), obtain the N top conformers (note, that the input is already clustered)
        # also get rid of duplicates in the indices
        rep_indices = []
        for solvent_key in self.settings.additional[_SC.FREE_ENERGY_SOLVENT_TAGS]:
            conf_indices = [tuple_conf[0] for tuple_conf in cluster_set]
            solvent_dGs = [
                float(tuple_conf[1].get_molecule().GetProp(solvent_key))
                for tuple_conf in cluster_set
            ]

            # sort list of global indices for this cluster according to their free energy for this solvent
            # note: from lowest (most negative) -> highest
            conf_indices_sorted = [
                idx for _, idx in sorted(zip(solvent_dGs, conf_indices))
            ]
            rep_indices = (
                rep_indices
                + conf_indices_sorted[
                    0 : min(
                        len(conf_indices),
                        self.settings.additional[_SC.TOP_N_PER_SOLVENT],
                    )
                ]
            )
        return list(set(rep_indices))

    def _cluster_conformers(self, conformers: List[Conformer]) -> List[Conformer]:
        # make sure the number of clusters specified and "N top per solvent" are not higher than the compound number
        n_clusters, top_n_per_solvent = self._get_nclusters_and_top_n(
            len_conformers=len(conformers)
        )

        # initialize K-means instance
        kmeans = KMeans(
            n_clusters=n_clusters,
            max_iter=self.settings.arguments.parameters[_SC.MAX_ITER],
            init="k-means++",
            n_init=10,
            tol=1e-04,
            random_state=0,
        )

        # generate dataframe with selected properties
        df_features = self._generate_feature_dataframe(conformers=conformers)

        # predict cluster and assign to conformer
        cluster_labels = kmeans.fit_predict(df_features)
        keep_indices = []
        for cluster_label in range(n_clusters):
            # keep the "global" index to select the appropriate conformers later
            cluster_set = [
                (i, conformers[i])
                for i in range(len(conformers))
                if cluster_labels[i] == cluster_label
            ]
            keep_indices = keep_indices + self._get_representative_conformers(
                cluster_set=cluster_set
            )
        return [conformers[i] for i in range(len(conformers)) if i in keep_indices]

    def execute(self):
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if len(enumeration.get_conformers()) == 0:
                    continue

                number_conformers_before = len(enumeration)

                # cluster conformers on the enumeration level
                clustered_conformers = self._cluster_conformers(
                    conformers=enumeration.get_conformers()
                )

                # add clustered conformers to enumeration
                enumeration.clear_conformers()
                for conf in clustered_conformers:
                    enumeration.add_conformer(conformer=conf, auto_update=True)
                number_conformers_after = len(enumeration)
                self._logger.log(
                    f"Clustered {number_conformers_before} into {number_conformers_after} conformers for enumeration {enumeration.get_index_string()}.",
                    _LE.INFO,
                )



from icolos.core.workflow_steps.step import StepBase
from pydantic import BaseModel
from openmm.app import PDBFile

import parmed
from openff.toolkit.topology import Molecule, Topology
from openff.toolkit.typing.engines.smirnoff import ForceField
from openff.toolkit.utils import get_data_file_path

from icolos.utils.enums.step_enums import StepOpenFFEnum
import os

# Note that this implementation leverages parmed for now, although will likely move over to the OpenFF Interchange tooling once stable
_SOFE = StepOpenFFEnum()

# TOOD: Eventually this step should be able to do a pdb2gmx job
class StepOFF2gmx(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

    def parametrise_mols(self, tmp_dir: str):
        """
        Generate parameters for each mol
        """
        # TODO: do we want to throw everything together or split the params into separate files?
        mols = [
            Molecule.from_smiles(smi)
            for smi in self._get_additional_setting(_SOFE.UNIQUE_MOLS)
        ]
        pdb_file = PDBFile(
            os.path.join(tmp_dir, self.data.generic.get_argument_by_extension("pdb"))
        )
        omm_topology = pdb_file.topology

        off_topology = Topology.from_openmm(omm_topology, unique_molecules=mols)

        forcefield = ForceField(self._get_additional_setting(_SOFE.FORCEFIELD))

        omm_system = forcefield.create_openmm_system(off_topology)

        if self._get_additional_setting(_SOFE.METHOD, _SOFE.PARMED) == _SOFE.PARMED:
            parmed_struct = parmed.openmm.load_topology(
                omm_topology, omm_system, pdb_file.positions
            )
            parmed_struct.save(os.path.join(tmp_dir, "MOL.top"), overwrite=True)
            parmed_struct.save(os.path.join(tmp_dir, "MOL.gro"), overwrite=True)

            # TODO: validate energy differences

        elif self._get_additional_setting(_SOFE.METHOD) == _SOFE.INTERCHANGE:
            raise NotImplementedError
        else:
            raise NotImplementedError

    def execute(self):
        """
        Builds a system and parametrise using OpenFF SAGE params, then convert to a GROMACS top/gro format for downstream simulation
        """
        tmp_dir = self._prepare_tmpdir()
        self.data.generic.write_out_all_files(tmp_dir)

        self.parametrise_mols(tmp_dir)

        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)


# If we want to build OpenFF params instead of gaff, we would need to make a call to a different parametrisation pipeline, then load the protein into a ParmEd System, combine with the OpenFF-built system and combine into a gro/top file.



from icolos.utils.general.files_paths import attach_root_path


# dictionary convenience functions
# ---------


def nested_get(dictionary: dict, keys: list, default=None):
    # recursively walk through nested dict, pull the value corresponding to the specified key(s)
    if not isinstance(keys, list):
        keys = [keys]
    if dictionary is None:
        return default
    if not keys:
        return dictionary
    return nested_get(dictionary.get(keys[0]), keys[1:], default)


def in_keys(dictionary: dict, keys: list) -> bool:
    if not isinstance(keys, list):
        keys = [keys]

    _dict = dictionary
    for key in keys:
        try:
            _dict = _dict[key]
        except KeyError:
            return False
    return True


# parsing "setup.py"
# ---------


def parse_setuppy():
    path = attach_root_path("setup.py")
    parsed_dict = {}
    with open(path, "r") as f:
        lines = f.readlines()
        for line in lines:
            if "name" in line:
                parsed_dict["name"] = line[line.find('"') + len('"') : line.rfind('"')]
            if "version" in line:
                parsed_dict["version"] = line[
                    line.find('"') + len('"') : line.rfind('"')
                ]
            if "license" in line:
                parsed_dict["license"] = line[
                    line.find('"') + len('"') : line.rfind('"')
                ]
            if "author" in line:
                parsed_dict["author"] = line[
                    line.find('"') + len('"') : line.rfind('"')
                ]
    return parsed_dict


def ensure_list(inp) -> list:
    return inp if isinstance(inp, list) else [inp]



from copy import deepcopy
from typing import List, Dict
from pydantic import BaseModel

from icolos.core.containers.compound import Enumeration, Compound
from icolos.utils.enums.step_enums import StepBaseEnum

_SBE = StepBaseEnum


class StepMerge(BaseModel):
    compounds: bool = True
    enumerations: bool = False
    merge_compounds_by: str = _SBE.INPUT_MERGE_BY_NAME
    merge_enumerations_by: str = _SBE.INPUT_MERGE_BY_ID


class InputMerger:
    def __init__(self, config: StepMerge):
        self.config = config

    def _group_enumerations(
        self, enumerations: List[Enumeration], by
    ) -> Dict[str, List[Enumeration]]:
        if by == _SBE.INPUT_MERGE_BY_SMILE:
            grouped = {enumeration.get_smile(): [] for enumeration in enumerations}
            for enum in enumerations:
                grouped[enum.get_smile()].append(enum)
        elif by == _SBE.INPUT_MERGE_BY_ID:
            grouped = {
                str(enumeration.get_enumeration_id()): []
                for enumeration in enumerations
            }
            for enum in enumerations:
                grouped[str(enum.get_enumeration_id())].append(enum)
        else:
            raise NotImplementedError
        return grouped

    def _group_compounds(
        self, compounds: List[Compound], by
    ) -> Dict[str, List[Compound]]:
        if by == _SBE.INPUT_MERGE_BY_NAME:
            names = {compound.get_name(): [] for compound in compounds}
            for compound in compounds:
                names[compound.get_name()].append(compound)
        elif by == _SBE.INPUT_MERGE_BY_ID:
            names = {str(compound.get_compound_number()): [] for compound in compounds}
            for compound in compounds:
                names[str(compound.get_compound_number())].append(compound)
        else:
            raise NotImplementedError
        return names

    def _merge_enumerations(
        self, enumerations: List[Enumeration], by
    ) -> List[Enumeration]:
        list_result = []

        # note that if it has been grouped by ID, the first (arbitrary) smile is used
        for _, enum_list in self._group_enumerations(enumerations, by).items():
            enum_combined = deepcopy(enum_list[0])
            enum_combined.clear_conformers()
            for enum in enum_list:
                enum_combined.add_conformers(
                    deepcopy(enum.get_conformers()), auto_update=False
                )
            list_result.append(enum_combined)
        return list_result

    def unroll_compounds(self, compounds: list) -> List[Compound]:
        list_buffer = []
        for ele in compounds:
            if isinstance(ele, list):
                list_buffer = list_buffer + self.unroll_compounds(ele)
            elif isinstance(ele, Compound):
                list_buffer.append(ele)
        return list_buffer

    def merge(self, compounds: List[Compound]) -> List[Compound]:
        list_result = []

        # if selected, combined compounds into one depending on the strategy
        if self.config.compounds:
            dict_grouped = self._group_compounds(
                compounds, self.config.merge_compounds_by
            )
            number = 0
            for name, compound_list in dict_grouped.items():
                # add the enumerations of all compounds together but do NOT auto-update yet (because enumerations might
                # also be merged later on)
                comp_combined = Compound(name=name, compound_number=number)
                for comp in compound_list:
                    comp_combined.add_enumerations(
                        deepcopy(comp.get_enumerations()), auto_update=False
                    )

                # as merging of enumerations only makes sense when there was a compound merge, keep
                # it on that indentation level
                if self.config.enumerations:
                    enumerations = self._merge_enumerations(
                        deepcopy(comp_combined.get_enumerations()),
                        self.config.merge_enumerations_by,
                    )
                    comp_combined.clear_enumerations()
                    comp_combined.add_enumerations(enumerations, auto_update=False)

                # now, rename the enumerations and conformers
                comp_combined.reset_all_ids()
                comp_combined.update_all_relations()

                list_result.append(comp_combined)
                number += 1
        return list_result



import os
from typing import List
from icolos.core.workflow_steps.step import StepBase
from icolos.utils.enums.step_enums import StepBaseEnum

_SBE = StepBaseEnum


class ConsoleColours:
    HEADER = "\033[95m"
    BLUE = "\033[94m"
    CYAN = "\033[96m"
    GREEN = "\033[92m"
    RED = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    ORANGE = "\033[0;33m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    BLINKING = "\33[5m"


def add_citation(
    step_type: str,
    workflow_step_types: List[str],
    citations: List[str],
    citation_string: str,
):
    if (
        any(
            [
                True
                for w_step_type in workflow_step_types
                if step_type.upper() in w_step_type
            ]
        )
        or step_type == "default"
    ):
        citations.append(citation_string)


def print_citations(steps: List[StepBase], logger=None):
    header = f"""{ConsoleColours.GREEN}
=====================================================================
ooooo   .oooooo.     .oooooo.   ooooo          .oooooo.    .oooooo..o 
`888'  d8P'  `Y8b   d8P'  `Y8b  `888'         d8P'  `Y8b  d8P'    `Y8 
 888  888          888      888  888         888      888 Y88bo.      
 888  888          888      888  888         888      888  `"Y8888o.  
 888  888          888      888  888         888      888      `"Y88b 
 888  `88b    ooo  `88b    d88'  888       o `88b    d88' oo     .d8P 
o888o  `Y8bood8P'   `Y8bood8P'  o888ooooood8  `Y8bood8P'  8""88888P'  
   ====================================================================={ConsoleColours.ENDC}\n"""
    writeout_lines = header.split("\n")
    writeout_lines.append(
        "If you publish work using Icolos, please consider citing the following papers, based on the workflow's steps...\n\n"
    )
    citations = []
    step_types = [step.type.upper() for step in steps]

    # add general Icolos citation
    add_citation(
        "default",
        step_types,
        citations,
        "\nIcolos: J. Harry Moore, Matthias R. Bauer, Jeff Guo, Atanas Patronov, Ola Engkvist and Christian Margreitter. Icolos: A workflow manager for structure based post-processing of de novo generated small molecules. https://doi.org/10.26434/chemrxiv-2022-sjcp3, https://github.com/MolecularAI/Icolos (2022)\n",
    )

    # add individual citations based on the step types used in this workflow (can be empty, of course)
    add_citation(
        "mdrun",
        step_types,
        citations,
        "\nGROMACS: https://doi.org/10.1016/j.softx.2015.06.001\nH.J.C. Berendsen, D. van der Spoel, and R. van Drunen, “GROMACS: A message-passing parallel molecular dynamics implementation,” Comp. Phys. Comm., 91 43–56 (1995)\nD. van der Spoel, E. Lindahl, B. Hess, G. Groenhof, A.E. Mark, and H.J.C. Berendsen, “GROMACS: Fast, Flexible and Free,” J. Comp. Chem., 26 1701–1718 (2005)\nH. Bekker, H.J.C. Berendsen, E.J. Dijkstra, S. Achterop, R. van Drunen, D. van der Spoel, A. Sijbers, and H. Keegstra et al., “Gromacs: A parallel computer for molecular dynamics simulations”; pp. 252–256 in Physics computing 92. Edited by R.A. de Groot and J. Nadrchal. World Scientific, Singapore, 1993",
    )
    add_citation(
        "pmx_",
        step_types,
        citations,
        "\nPMX: Vytautas Gapsys, Servaas Michielssens, Daniel Seeliger and Bert L. de Groot. Accurate and Rigorous Large Scale Mutation Free Energy Prediction. Angewandte Chemie Int. Ed. 55: 7364-7368(2016)\nVytautas Gapsys, Servaas Michielssens, Daniel Seeliger, and Bert L. de Groot. pmx: Automated protein structure and topology generation for alchemical perturbations. J. Comput. Chem. 36:348-354 (2015\n",
    )
    add_citation(
        "gmx_MMPBSA",
        step_types,
        citations,
        "\ngmx_MMPBSA: Valdés-Tresanco, M.S., Valdés-Tresanco, M.E., Valiente, P.A. and Moreno E. gmx_MMPBSA: A New Tool to Perform End-State Free Energy Calculations with GROMACS. Journal of Chemical Theory and Computation, 2021 17 (10), 6281-6291\n",
    )
    add_citation(
        "fep",
        step_types,
        citations,
        "\nFEP+: R. Abel, L. Wang, E.D. Harder, B.J. Berne, R.A. Friesner. Advancing Drug Discovery through Enhanced Free Energy Calculations. Acc. Chem. Res., 2017 50 (7), 1625-1632\n",
    )
    add_citation(
        _SBE.STEP_KALLISTO,
        step_types,
        citations,
        "\nKallisto: Caldeweyher, E., Kallisto: A command-line interface to simplify computational modelling and the generation of atomic features. Journal of Open Source Software, 6(60), 3050, https://doi.org/10.21105/joss.03050. 2021\n",
    )
    add_citation(
        _SBE.STEP_JAZZY,
        step_types,
        citations,
        "\nJazzy: Caldeweyher, E., Ghiandoni, G.M., https://github.com/AstraZeneca/jazzy <CITATION MISSING>\n",
    )
    add_citation(
        _SBE.STEP_TURBOMOLE,
        step_types,
        citations,
        "\nTURBOMOLE: Modular program suite for ab initio quantum-chemical and condensed-matter simulations; J. Chem. Phys. 152, 184107 (2020); https://doi.org/10.1063/5.0004635\n",
    )
    add_citation(
        _SBE.STEP_ESP_SIM,
        step_types,
        citations,
        "\nESPSim: https://doi.org/10.26434/chemrxiv-2021-sqvv9-v3\n",
    )
    add_citation(
        _SBE.STEP_AUTODOCKVINA_DOCKING,
        step_types,
        citations,
        "\nAutoDock Vina: Eberhardt, J., Santos-Martins, D., Tillack, A.F., Forli, S. (2021). AutoDock Vina 1.2.0: New Docking Methods, Expanded Force Field, and Python Bindings. Journal of Chemical Information and Modeling.\nTrott, O., & Olson, A. J. (2010). AutoDock Vina: improving the speed and accuracy of docking with a new scoring function, efficient optimization, and multithreading. Journal of computational chemistry, 31(2), 455-461.\n",
    )
    add_citation(
        _SBE.STEP_XTB,
        step_types,
        citations,
        "\nGFN2-xTB: C. Bannwarth, S. Ehlert, S. Grimme. GFN2-xTB—An Accurate and Broadly Parametrized Self-Consistent Tight-Binding Quantum Chemical Method with Multipole Electrostatics and Density-Dependent Dispersion Contributions.\nJ. Chem. Theory Comput. 2019, 15 (3), 1652–1671 DOI: 10.1021/acs.jctc.8b01176\n",
    )
    add_citation(
        _SBE.STEP_OMEGA,
        step_types,
        citations,
        "\nOMEGA: P.C.D Hawkins, A.G. Skillman, G.L. Warren, B.A. Ellingson, M.T. Stahl. Conformer Generation with OMEGA: Algorithm and Validation Using High Quality Structures from the Protein Databank and the Cambridge Structural Database. J. Chem. Inf. Model. 2010, 50, 572-584.\n",
    )
    add_citation(
        _SBE.STEP_COSMO,
        step_types,
        citations,
        "\nCOSMOtherm, Version C3.0, Release 17.01; COSMOlogic GmbH & Co. KG, http://www.cosmologic.de\n",
    )
    add_citation(
        _SBE.STEP_SHAEP,
        step_types,
        citations,
        "\nShaEP: M.J. Vainio, J.S. Puranen, M.S. Johnson. ShaEP: Molecular Overlay Based on Shape and Electrostatic Potential. J. Chem. Inf. Model. 49, 492-502 (2009).\n",
    )
    add_citation(
        _SBE.STEP_PANTHER,
        step_types,
        citations,
        "\nS.P. Niinivehmas, K. Salokas, S. Lätti, H. Raunio, O.T. Pentikäinen. Ultrafast protein structure-based virtual screening with Panther. J. Comput. Aided. Mol. Des. 29, 989–1006. doi: 10.1007/s10822-015-9870-3 (2015).\n",
    )
    add_citation(
        _SBE.STEP_MDPOCKET,
        step_types,
        citations,
        "\nMDpocket: P. Schmidtke, A. Bidon-Chanal, F.J. Luque, X. Barril. MDpocket: open-source cavity detection and characterization on molecular dynamics trajectories. Bioinformatics, Volume 27, Issue 23, Pages 3276–3285. https://doi.org/10.1093/bioinformatics/btr550 (2011)\n",
    )
    add_citation(
        _SBE.STEP_GOLD_DOCKING,
        step_types,
        citations,
        "\nG. Jones, P. Willett, R.C. Glen, A.R. Leach, R. Taylor. Development and Validation of a Genetic Algorithm for Flexible Docking. J. Mol. Biol., 267, 727-748. doi: 10.1006/jmbi.1996.0897 (1997).\n",
    )
    add_citation(
        _SBE.STEP_CREST,
        step_types,
        citations,
        "\nCREST: P. Pracht, F. Bohle, S. Grimme. Phys. Chem. Chem. Phys., 22, 7169-7192. doi: 10.1039/C9CP06869D (2020).\n",
    )
    add_citation(
        _SBE.STEP_GLIDE,
        step_types,
        citations,
        "\nGLIDE: R.A. Friesner, R.B. Murphy, M.P. Repasky, L.L. Frye, J.R. Greenwood, T.A. Halgren, P.C. Sanschagrin, D.T. Mainz. Extra Precision Glide: Docking and Scoring Incorporating a Model of Hydrophobic Enclosure for Protein-Ligand Complexes. J. Med. Chem., 49, 6177–6196 (2006).\n",
    )
    add_citation(
        _SBE.STEP_LIGPREP,
        step_types,
        citations,
        "\nSchrödinger Release 2021-4: LigPrep, Schrödinger, LLC, New York, NY (2021).\n",
    )
    add_citation(
        _SBE.STEP_MACROMODEL,
        step_types,
        citations,
        "\nSchrödinger Release 2021-4: MacroModel, Schrödinger, LLC, New York, NY (2021).\n",
    )
    add_citation(
        _SBE.STEP_PRIME,
        step_types,
        citations,
        "\nM.P. Jacobson, D.L. Pincus, C.S. Rapp, T.J.F. Day, B. Honig, D.E. Shaw, R.A. Friesner. A Hierarchical Approach to All-Atom Protein Loop Prediction. Proteins: Structure, Function and Bioinformatics, 55, 351-367 (2004).\n",
    )

    citations = list(set(citations))
    for c in citations:
        writeout_lines.append(c)
    writeout_lines.append("\n")
    writeout_lines.append(
        "=====================================================================\n"
    )

    for line in writeout_lines:
        try:
            width = os.get_terminal_size().columns
        except (OSError, AttributeError):
            width = 150

        print(line.center(width))



from icolos.utils.enums.step_enums import StepBaseEnum, StepGromacsEnum
from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.utils.execute_external.gromacs import GromacsExecutor
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase

_GE = GromacsEnum()
_SGE = StepGromacsEnum()
_SBE = StepBaseEnum


class StepGMXDoDSSP(StepGromacsBase, BaseModel):
    """
    Wrapper for gmx do_dssp binary, which in turn wraps dssp
    returns secondary structure data for the provided gromacs trajectory
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)
        self._check_backend_availability()

    def execute(self):
        tmp_dir = self._make_tmpdir()
        self.write_input_files(tmp_dir)

        structure_file = self.data.generic.get_argument_by_extension(_SGE.FIELD_KEY_TPR)
        traj_file = self.data.generic.get_argument_by_extension(_SGE.FIELD_KEY_XTC)
        arguments = self._parse_arguments(
            flag_dict={"-f": traj_file, "-s": structure_file, "-ssdump": "info.dat"}
        )

        pipe_input = (
            self.construct_pipe_arguments(
                tmp_dir, self.settings.additional[_SBE.PIPE_INPUT]
            )
            if _SBE.PIPE_INPUT in self.settings.additional.keys()
            and self.settings.additional[_SBE.PIPE_INPUT] is not None
            else None
        )

        result = self._backend_executor.execute(
            command=_GE.DO_DSSP,
            arguments=arguments,
            location=tmp_dir,
            pipe_input=pipe_input,
        )
        for line in result.stdout.split("\n"):
            self._logger_blank.log(line, _LE.DEBUG)
        self._logger.log(
            f"Completed execution for {self.step_id} successfully", _LE.INFO
        )

        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



class ExecutorEnum:

    RUNTIME_GLOBAL_VARIABLE_WORKDIR = "work_dir"
    RUNTIME_GLOBAL_VARIABLE_ENTRYPOINTDIR = "entrypoint_dir"
    RUNTIME_GLOBAL_VARIABLE_CONFIGDIR = "config_dir"
    RUNTIME_GLOBAL_VARIABLE_PACKAGEDIR = "package_dir"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")



class CompoundTagsEnum:

    CONFORMER_ENERGY_TAG = "conformer_energy"
    FORMAL_CHARGE_TAG = "formal_charge"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class CompoundContainerEnum:

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class EnumerationContainerEnum:

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class ConformerContainerEnum:

    EXTRA_DATA_COSMOFILE = "cosmo_file"
    EXTRA_DATA_COORDFILE = "coord_file"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")



from icolos.core.step_dispatch.dispatcher import StepDispatcher
from icolos.core.workflow_steps.autodockvina.docking import StepAutoDockVina
from icolos.core.workflow_steps.autodockvina.target_preparation import (
    StepAutoDockVinaTargetPreparation,
)
from icolos.core.workflow_steps.calculation.jazzy import StepJazzy
from icolos.core.workflow_steps.calculation.kallisto import StepKallisto
from icolos.core.workflow_steps.calculation.lomap import StepLomap
from icolos.core.workflow_steps.ccdc.docking import StepGold
from icolos.core.workflow_steps.calculation.esp_sim import StepEspSim
from icolos.core.workflow_steps.calculation.feature_counter import StepFeatureCounter
from icolos.core.workflow_steps.gromacs.do_dssp import StepGMXDoDSSP
from icolos.core.workflow_steps.gromacs.mmpbsa import StepGMXmmpbsa
from icolos.core.workflow_steps.fpocket.mdpocket import StepMDpocket
from icolos.core.workflow_steps.gromacs.trajcat import StepGMXTrjcat
from icolos.core.workflow_steps.io.data_manipulation import StepDataManipulation
from icolos.core.workflow_steps.schrodinger.fep_absolute import (
    StepSchrodingerAbsoluteFEP,
)
from icolos.core.workflow_steps.schrodinger.fep_analysis import StepFepPlusAnalysis
from icolos.core.workflow_steps.schrodinger.protein_interaction import (
    StepProteinInteraction,
)
from icolos.core.workflow_steps.structure_prediction.pdb_fixer import StepPdbFixer
from icolos.core.workflow_steps.gromacs import *
from icolos.core.workflow_steps.calculation.boltzmann_weighting import (
    StepBoltzmannWeighting,
)
from icolos.core.workflow_steps.calculation.rmsd import StepRMSD
from icolos.core.workflow_steps.schrodinger import *
from icolos.core.workflow_steps.calculation.cosmo import StepCosmo
from icolos.core.workflow_steps.calculation.turbomole import StepTurbomole
from icolos.core.workflow_steps.confgen.crest import StepCREST
from icolos.core.workflow_steps.pmx import *
from icolos.core.workflow_steps.confgen.omega import StepOmega
from icolos.core.workflow_steps.confgen.xtb import StepXTB
from icolos.core.workflow_steps.io.embedder import StepEmbedding
from icolos.core.workflow_steps.io.initialize_compound import StepInitializeCompound
from icolos.core.workflow_steps.prediction.predictor import StepPredictor
from icolos.core.workflow_steps.prediction.model_building import StepModelBuilder
from icolos.core.workflow_steps.calculation.clustering import StepClustering
from icolos.core.workflow_steps.calculation.rms_filter import StepRMSFilter
from icolos.core.workflow_steps.calculation.panther import StepPanther
from icolos.core.workflow_steps.calculation.shaep import StepShaep
from icolos.core.workflow_steps.structure_prediction.peptide_embedder import (
    StepPeptideEmbedder,
)

from icolos.core.workflow_steps.active_learning.prospective_reinvent import (
    StepProspectiveREINVENT,
)

from icolos.core.workflow_steps.structure_prediction.dssp import StepDSSP
from icolos.utils.enums.step_enums import StepBaseEnum


_SBE = StepBaseEnum


class StepInitializationEnum:

    STEP_INIT_DICT = {
        _SBE.STEP_CREST: StepCREST,
        _SBE.STEP_OMEGA: StepOmega,
        _SBE.STEP_XTB: StepXTB,
        _SBE.STEP_MACROMODEL: StepMacromodel,
        _SBE.STEP_TURBOMOLE: StepTurbomole,
        _SBE.STEP_COSMO: StepCosmo,
        _SBE.STEP_INITIALIZATION: StepInitializeCompound,
        _SBE.STEP_EMBEDDING: StepEmbedding,
        _SBE.STEP_PREDICTION: StepPredictor,
        _SBE.STEP_MODEL_BUILDING: StepModelBuilder,
        _SBE.STEP_BOLTZMANN_WEIGHTING: StepBoltzmannWeighting,
        _SBE.STEP_PRIME: StepPrime,
        _SBE.STEP_DESMOND: StepDesmondExec,
        _SBE.STEP_DESMOND_SETUP: StepDesmondSetup,
        _SBE.STEP_CLUSTERING: StepClustering,
        _SBE.STEP_RMSFILTER: StepRMSFilter,
        _SBE.STEP_PANTHER: StepPanther,
        _SBE.STEP_KALLISTO: StepKallisto,
        _SBE.STEP_JAZZY: StepJazzy,
        _SBE.STEP_SHAEP: StepShaep,
        _SBE.STEP_PDB2GMX: StepGMXPdb2gmx,
        _SBE.STEP_EDITCONF: StepGMXEditConf,
        _SBE.STEP_SOLVATE: StepGMXSolvate,
        _SBE.STEP_GENION: StepGMXGenion,
        _SBE.STEP_GROMPP: StepGMXGrompp,
        _SBE.STEP_MDRUN: StepGMXMDrun,
        _SBE.STEP_TRJCONV: StepGMXTrjconv,
        _SBE.STEP_TRJCAT: StepGMXTrjcat,
        _SBE.STEP_CLUSTER: StepGMXCluster,
        _SBE.STEP_DO_DSSP: StepGMXDoDSSP,
        _SBE.STEP_GMX_RMSD: StepGMXrmsd,
        _SBE.STEP_LIGPREP: StepLigprep,
        _SBE.STEP_GLIDE: StepGlide,
        _SBE.STEP_FEP_PLUS_SETUP: StepFepPlusSetup,
        _SBE.STEP_FEP_PLUS_EXEC: StepFepPlusExec,
        _SBE.STEP_FEP_ABSOLUTE: StepSchrodingerAbsoluteFEP,
        _SBE.STEP_FEP_PLUS_ANALYSIS: StepFepPlusAnalysis,
        _SBE.STEP_PREPWIZARD: StepPrepwizard,
        _SBE.STEP_RESIDUE_SCANNING: StepResidueScanning,
        _SBE.STEP_MDPOCKET: StepMDpocket,
        _SBE.STEP_PEPTIDE_EMBEDDER: StepPeptideEmbedder,
        _SBE.STEP_PDB_FIXER: StepPdbFixer,
        _SBE.STEP_GMX_MMPBSA: StepGMXmmpbsa,
        _SBE.STEP_TS_CLUSTER: StepClusterTS,
        _SBE.STEP_DSSP: StepDSSP,
        _SBE.STEP_RMSD: StepRMSD,
        _SBE.STEP_DATA_MANIPULATION: StepDataManipulation,
        _SBE.STEP_PMX_ASSEMBLE_SYSTEMS: StepPMXAssembleSystems,
        _SBE.STEP_PMX_ATOMMAPPING: StepPMXatomMapping,
        _SBE.STEP_PMX_ABFE_SETUP: StepPMXabfe,
        _SBE.STEP_PMX_DOUBLEBOX: StepPMXdoublebox,
        _SBE.STEP_PMX_LIGANDHYBRID: StepPMXligandHybrid,
        _SBE.STEP_PMX_BOX_WATER_IONS: StepPMXBoxWaterIons,
        _SBE.STEP_PMX_SETUP: StepPMXSetup,
        _SBE.STEP_PMX_PREPARE_TRANSITIONS: StepPMXPrepareTransitions,
        _SBE.STEP_PMX_PREPARE_SIMULATIONS: StepPMXPrepareSimulations,
        _SBE.STEP_PMX_RUN_ANALYSIS: StepPMXRunAnalysis,
        _SBE.STEP_PMX_MUTATE: StepPMXmutate,
        _SBE.STEP_PMX_GENTOP: StepPMXgentop,
        _SBE.STEP_PMX_GENLIB: StepPMXgenlib,
        _SBE.STEP_FEATURE_COUNTER: StepFeatureCounter,
        _SBE.STEP_AUTODOCKVINA_DOCKING: StepAutoDockVina,
        _SBE.STEP_AUTODOCKVINA_TARGET_PREPARATION: StepAutoDockVinaTargetPreparation,
        _SBE.STEP_GOLD_DOCKING: StepGold,
        _SBE.STEP_PMX_RUN_SIMULATIONS: StepPMXRunSimulations,
        _SBE.STEP_DISPATCHER: StepDispatcher,
        _SBE.STEP_ESP_SIM: StepEspSim,
        _SBE.STEP_LOMAP: StepLomap,
        _SBE.STEP_PROSPECTIVE_REINVENT: StepProspectiveREINVENT,
        _SBE.STEP_PROTEIN_INTERACTIONS: StepProteinInteraction,
    }



import os
import sys

import pandas as pd
import argparse

import rdkit.Chem as Chem

from icolos.core.step_utils.rdkit_utils import to_smiles


def main():

    # get the input parameters and parse them
    parser = argparse.ArgumentParser(
        description="Implements simple translator taking an SDF file and spitting out SMILES."
    )
    parser.add_argument("-sdf", type=str, default=None, help="A path a SDF file.")
    parser.add_argument(
        "-smi",
        type=str,
        default=None,
        required=False,
        help="A path an output text file.",
    )
    parser.add_argument(
        "-csv",
        type=str,
        default=None,
        required=False,
        help="A path an output CSV file.",
    )
    parser.add_argument(
        "-keep_stereo",
        action="store_true",
        help="If set, exported SMILES contain stereo-information.",
    )
    parser.add_argument(
        "-tags2columns",
        type=str,
        nargs="+",
        default=None,
        required=False,
        help="A list of strings for which tags should be transformed into columns.",
    )
    args = parser.parse_args()

    if args.sdf is None or not os.path.isfile(args.sdf):
        raise Exception(
            'Parameter "-sdf" must be a relative or absolute path to valid sdf file.'
        )
    if args.smi is None and args.csv is None:
        raise Exception(
            'At least one of the "-smi" or "-csv" output paths must be set.'
        )

    molecules = []
    for mol in Chem.SDMolSupplier(args.sdf):
        if mol is None:
            continue
        molecules.append(mol)

    # write out
    # ---------
    if args.smi is not None:
        with open(args.smi, "w") as smi_file:
            for mol in molecules:
                smi_file.write(to_smiles(mol, isomericSmiles=args.keep_stereo) + "\n")

    if args.csv is not None:
        data_buffer = []
        columns = ["Name", "SMILES"]
        tags2columns = []
        if args.tags2columns is not None:
            tags2columns = args.tags2columns
            columns = columns + tags2columns
        for mol in molecules:
            # add default columns for this row
            row = [
                mol.GetProp("_Name"),
                to_smiles(mol, isomericSmiles=args.keep_stereo),
            ]

            # add selected columns for this row (if specified)
            for tag in tags2columns:
                try:
                    row.append(mol.GetProp(tag))
                except KeyError:
                    row.append(None)

            data_buffer.append(row)
        df_writeout = pd.DataFrame(data_buffer, columns=columns)
        df_writeout.to_csv(
            path_or_buf=args.csv,
            sep=",",
            na_rep="",
            header=True,
            index=False,
            mode="w",
            quoting=None,
        )


if __name__ == "__main__":
    ret = main()
    sys.exit(0)



import pickle
from copy import deepcopy

import numpy as np
from typing import List

from pydantic import BaseModel
from rdkit import Chem

from icolos.utils.general.icolos_exceptions import StepFailed, get_exception_message
from icolos.utils.enums.step_enums import StepPredictorEnum
from icolos.core.workflow_steps.io.base import StepIOBase
from icolos.core.workflow_steps.step import _LE

from icolos.utils.general.convenience_functions import *

_SPE = StepPredictorEnum()


class StepPredictor(StepIOBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

    @classmethod
    def _load_scikit_model(cls, model_path: str):
        with open(model_path, "rb") as f:
            scikit_model = pickle.load(f)
            return scikit_model

    def _get_feature_values(
        self, conformer: Chem.Mol, feature_names: List[str]
    ) -> np.ndarray:
        list_values = []
        for feature in feature_names:
            try:
                list_values.append(float(conformer.GetProp(feature)))
            except KeyError as e:
                self._logger.log(
                    f"Could not find feature / property, error message: {get_exception_message(e)}",
                    _LE.ERROR,
                )
                raise e

        # cast list to 2D array
        return np.array([list_values])

    def execute(self):
        # get parameters
        parameters = deepcopy(self.settings.additional)
        model_path = nested_get(parameters, _SPE.MODEL_PATH, default=None)
        feature_names = nested_get(parameters, _SPE.FEATURES, default=None)
        name_predicted = nested_get(
            parameters, _SPE.NAME_PREDICTED, default=_SPE.NAME_PREDICTED_DEFAULT
        )

        # check parameters; model_path and features are mandatory
        if model_path is None or feature_names is None:
            message = f"Parameters {_SPE.MODEL_PATH} (path to model) and {_SPE.FEATURES} (list with features) have to be set - abort."
            self._logger.log(message, _LE.ERROR)
            raise StepFailed(message)
        if name_predicted == _SPE.NAME_PREDICTED_DEFAULT:
            self._logger.log(
                f"Name of predicted property not specified, using default value {_SPE.NAME_PREDICTED_DEFAULT} instead (not recommended).",
                _LE.WARNING,
            )

        # load model from file and predict endpoint
        model = self._load_scikit_model(model_path=model_path)
        predicted = 0
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                for conformer in enumeration.get_conformers():
                    if not self._input_object_valid(conformer):
                        continue

                    f_values = self._get_feature_values(
                        conformer=conformer.get_molecule(), feature_names=feature_names
                    )
                    conformer.get_molecule().SetProp(
                        name_predicted, str(model.predict(X=f_values)[0])
                    )
                    predicted += 1
        self._logger.log(
            f"Predicted {name_predicted} for {predicted} conformers in {len(self.get_compounds())} compounds.",
            _LE.INFO,
        )



from typing import List
from icolos.core.workflow_steps.step import StepBase
from icolos.utils.execute_external.execute import Executor
from pydantic import BaseModel
from icolos.utils.enums.step_enums import StepDSSPEnum
from icolos.utils.enums.program_parameters import DSSPEnum
import os


_SDE = StepDSSPEnum()
_DE = DSSPEnum()


class StepDSSP(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(executor=Executor)

    def _construct_arguments(self, tmp_dir: str, file: str) -> List:
        args = []
        for flag in self.settings.arguments.flags:
            args.append(flag)
        for key, value in self.settings.arguments.parameters.items():
            args.append(key)
            args.append(value)

        # set the input and output files
        args.append(file)
        output = f"dssp_output_{file.split('.')[0]}.txt"
        args.append(output)
        return args

    def _parse_output(self, tmp_dir: str) -> None:
        for file in [f for f in os.listdir(tmp_dir) if f.endswith("txt")]:
            with open(os.path.join(tmp_dir, file), "r") as f:
                self._add_data_to_generic(file, f.read())

    def execute(self):
        """
        Executes dssp on a set of input structures
        """

        tmp_dir = self._make_tmpdir()
        print(tmp_dir)
        self.data.generic.write_out_all_files(tmp_dir)

        file_list = self.data.generic.get_file_names_by_extension(ext="pdb")

        for file in file_list:
            arguments = self._construct_arguments(tmp_dir, file)
            self._backend_executor.execute(
                command=_DE.MKDSSP, arguments=arguments, check=True, location=tmp_dir
            )

        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.utils.execute_external.execute import ExecutorBase

_GE = GromacsEnum()


class GromacsExecutor(ExecutorBase):
    def __init__(self, prefix_execution=None, binary_location=None):
        super().__init__(
            prefix_execution=prefix_execution, binary_location=binary_location
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        gmx_commands = [
            _GE.EDITCONF,
            _GE.GENION,
            _GE.GROMPP,
            _GE.SOLVATE,
            _GE.MDRUN,
            _GE.MPI_MDRUN,
            _GE.PDB2GMX,
            _GE.MAKE_NDX,
            _GE.GENRESTR,
            _GE.TRJCONV,
            _GE.TRJCAT,
            _GE.CLUSTER,
            _GE.MMPBSA,
            _GE.DO_DSSP,
            _GE.RMS,
        ]

        if not any([cmd in command for cmd in gmx_commands]):
            raise ValueError(
                "Command must be present in internal list of GROMACS executables"
            )
        # handle for dealing with programs that want interactive input
        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=location,
            pipe_input=pipe_input,
        )

    def is_available(self):
        try:
            result = self.execute(command=_GE.PDB2GMX, arguments=[], check=False)
            if _GE.PDB2GMX_FAIL_ID_STRING in result.stderr:
                return True
            return False
        except Exception as e:
            return False



from copy import deepcopy
from typing import List
from icolos.core.containers.generic import GenericData
from icolos.core.step_utils.structconvert import StructConvert
from icolos.core.workflow_steps.schrodinger.base import StepSchrodingerBase
from icolos.utils.enums.program_parameters import (
    FepPlusEnum,
    SchrodingerExecutablesEnum,
)
from icolos.utils.enums.step_enums import StepBaseEnum, StepFepPlusEnum, StepGlideEnum
from icolos.utils.execute_external.fep_plus import FepPlusExecutor
from rdkit.Chem import SDMolSupplier
from icolos.utils.execute_external.schrodinger import SchrodingerExecutor
from icolos.core.workflow_steps.step import _LE
import os
from pydantic import BaseModel
from rdkit.Chem import SDWriter

_SFE = StepFepPlusEnum()
_FE = FepPlusEnum()
_SEE = SchrodingerExecutablesEnum()
_SBE = StepBaseEnum
_SGE = StepGlideEnum()


class StepFepPlusSetup(StepSchrodingerBase, BaseModel):
    """
    Construct and analyse perturbation map for set of congeneric ligands
    Supports extracting structures from poseviewer or pdb files
    """

    _schrodinger_executor: SchrodingerExecutor = None
    _converter: StructConvert = None

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=FepPlusExecutor)
        self._check_backend_availability()

        self._schrodinger_executor = SchrodingerExecutor(
            prefix_execution=self.execution.prefix_execution,
            binary_location=self.execution.binary_location,
        )
        self._converter = StructConvert(
            prefix_execution=self.execution.prefix_execution,
            binary_location=self.execution.binary_location,
        )

    def _extract_receptor_from_pv(self, tmp_dir, input_file: str = _SFE.RECEPTOR_MAEGZ):
        # run split_structure.py to obtain the receptor_structure
        self._logger.log(f"Extracting receptor from structure.", _LE.INFO)
        self._schrodinger_executor.execute(
            command=_SEE.STRUCT_SPLIT,
            arguments=[
                "-m",
                "pdb",
                "-many_files",
                os.path.join(tmp_dir, input_file),
                f"{_SFE.STRUCT_SPLIT_BASE}.mae",
            ],
            check=True,
            location=tmp_dir,
        )

        # get rid of the original receptor structure now we have the new one
        os.remove(os.path.join(tmp_dir, _SFE.RECEPTOR_MAEGZ))

    def _write_receptor_from_pv(self, tmp_dir):
        # Handles writing the receptor structure to tmpdir, either from a poseviewer file, or a provided receptor
        # take the first poseviewer file it can find and split the stricure, take only the receptor
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                for conformer in enumeration.get_conformers():
                    if (
                        _SGE.GLIDE_POSEVIEWER_FILE_KEY
                        in conformer.get_extra_data().keys()
                    ):
                        with open(
                            os.path.join(tmp_dir, _SFE.RECEPTOR_MAEGZ), "wb"
                        ) as f:
                            f.write(
                                conformer.get_extra_data()[
                                    _SGE.GLIDE_POSEVIEWER_FILE_KEY
                                ]
                            )
                            break
        if _SFE.RECEPTOR_MAEGZ in os.listdir(tmp_dir):
            self._logger.log(
                f"Writing poseviewer file to temporary directory.", _LE.INFO
            )
            self._extract_receptor_from_pv(tmp_dir)
        elif self.data.generic.get_files_by_extension("pdb"):
            # a pdb file was loaded to generic data, use this as the receptor structure
            self.data.generic.get_argument_by_extension(
                "pdb", rtn_file_object=True
            ).write(os.path.join(tmp_dir, "receptor.pdb"), join=False)

            self._logger.log(
                "Converting provided pdb receptor structure to mae", _LE.DEBUG
            )
            self._converter.convert(
                os.path.join(tmp_dir, "receptor.pdb"),
                os.path.join(tmp_dir, f"{_SFE.STRUCT_SPLIT_BASE}_receptor1.mae"),
            )
            os.remove(os.path.join(tmp_dir, "receptor.pdb"))

        else:
            self._logger.log(
                "No poseviewer file was found attached to any of the conformers, and no PDB receptor file was specified - this must be set in the docking step",
                _LE.ERROR,
            )
            raise FileNotFoundError

    def _check_xray_structure(self, compound_number):
        # check to see if an xray structure has been provided for that compound
        if _SFE.XRAY_STRUCTURES in self.settings.additional.keys():
            if isinstance(self.settings.additional[_SFE.XRAY_STRUCTURES], dict):
                if (
                    compound_number
                    in self.settings.additional[_SFE.XRAY_STRUCTURES].keys()
                ):
                    return True, _FE.DICT
            elif os.path.isdir(self.settings.additional[_SFE.XRAY_STRUCTURES]):
                if os.path.isfile(
                    os.path.join(
                        self.settings.additional[_SFE.XRAY_STRUCTURES],
                        f"{compound_number}.pdb",
                    )
                ):
                    return True, _FE.PATH
        return False, None

    def _rename_sdf(self, path, comp_num):
        with open(path, "r") as f:
            lines = f.readlines()[1:]
        new_lines = [f"{comp_num}:0:0\n"]
        for line in lines:
            new_lines.append(line)
        self._remove_temporary(path)
        with open(path, "w") as f:
            f.writelines(new_lines)

    def _extract_ligand_from_pdb(self, tmp_dir: str, comp_num: int, type: str):
        # if ligand poses have been provided from xray structures, extract just the ligand
        self._logger.log(
            f"Extracting ligand from provided Xray structure for compound {comp_num}",
            _LE.DEBUG,
        )
        if type == _FE.DICT:
            file_path = self.settings.additional[_SFE.XRAY_STRUCTURES[comp_num]]
        else:
            file_path = os.path.join(
                self.settings.additional[_SFE.XRAY_STRUCTURES], f"{comp_num}.pdb"
            )
        if not os.path.isfile(file_path):
            raise FileNotFoundError(
                "The provided path to the xray structure does not exist or is not accessible"
            )
        self._schrodinger_executor.execute(
            command=_SEE.STRUCT_SPLIT,
            arguments=["-m", "pdb", "-many_files", file_path, f"{_SFE.XRAY_SPLIT}.sdf"],
            check=True,
            location=tmp_dir,
        )
        # remove everything apart from the ligand sdf which is concatenated later
        lig_found = False
        for file in os.listdir(tmp_dir):
            idx = file.split("/")[-1]
            if idx.startswith(_SFE.XRAY_SPLIT):
                if "ligand" in idx:
                    # need to modify the name from the standard that Schrodinger provides
                    self._rename_sdf(os.path.join(tmp_dir, file), comp_num)
                    mols = SDMolSupplier(os.path.join(tmp_dir, file))

                    data = mols[0]
                    lig_found = True
                    self._remove_temporary(os.path.join(tmp_dir, file))
                else:
                    self._remove_temporary(os.path.join(tmp_dir, file))
        if lig_found:
            return data

    def _write_input_files(self, tmp_dir):
        # write receptor structure to tmpdir, either from poseviewer or provided pdb file
        self._write_receptor_from_pv(tmp_dir)

        # write out all conformers present in self.data.compounds to a single sdf file.
        writer = SDWriter(os.path.join(tmp_dir, "concatenated.sdf"))
        for compound in self.get_compounds():
            # If an xray pose is provided, use this
            flag, type = self._check_xray_structure(compound.get_compound_number())
            if flag is True:
                self._logger.log(
                    "Found Xray structure for the ligand - using this in preference to a docking pose",
                    _LE.DEBUG,
                )
                mol = self._extract_ligand_from_pdb(
                    tmp_dir, compound.get_compound_number(), type
                )
                writer.write(mol)
            else:
                # use the docked conformer
                for enumeration in compound.get_enumerations():
                    for conformer in enumeration.get_conformers():
                        mol = deepcopy(conformer.get_molecule())
                        mol.SetProp("_Name", conformer.get_compound_name())
                        writer.write(mol)

    def _parse_arguments(self, io_dict: dict) -> List[str]:
        arguments = []
        for key in self.settings.arguments.parameters.keys():
            arguments.append(key)
            arguments.append(str(self.settings.arguments.parameters[key]))
        for flag in self.settings.arguments.flags:
            arguments.append(str(flag))
        for key, value in io_dict.items():
            arguments.append(key)
            arguments.append(value)
        return arguments

    def _get_structcat_args(
        self, tmp_dir: str, out_file_type: str, outfile: str
    ) -> List[str]:
        arguments = [
            f"{_SEE.STRUCTCAT_I}mae",
            os.path.join(tmp_dir, f"{_SFE.STRUCT_SPLIT_BASE}_receptor1.mae"),
            f"{_SEE.STRUCTCAT_I}sd",
        ]

        for file in os.listdir(tmp_dir):
            if file.endswith("sdf"):
                arguments.append(os.path.join(tmp_dir, file))
        arguments.append(f"{_SEE.STRUCTCAT_O}{out_file_type}")
        arguments.append(os.path.join(tmp_dir, outfile))
        return arguments

    def _concatenate_pv_files(self, tmp_dir: str):
        # create a poseviewer-formatted file with receptor structure, then docked ligand poses
        arguments = self._get_structcat_args(
            tmp_dir=tmp_dir, out_file_type="mae", outfile=_SFE.STRUCTCAT_MAEGZ_OUTFILE
        )
        self._schrodinger_executor.execute(
            command=_SEE.STRUCTCAT, arguments=arguments, check=True
        )

    def _analyse_map(self, tmp_dir):
        """run fmp_stats program to analyse map - generate node similarities etc"""
        result = self._schrodinger_executor.execute(
            command=_SEE.FMP_STATS,
            arguments=["out.fmp", "-f"],
            check=True,
            location=tmp_dir,
        )
        log_lines = []
        for line in str(result.stdout).split("\n"):
            self._logger_blank.log(line, _LE.INFO)
            log_lines.append(line + "\n")

        self.data.generic.add_file(
            GenericData(file_name="fep_mapper.log", file_data=log_lines)
        )

    def _parse_output(self, tmp_dir: str):
        # needs to retrieve the edge and fmp files produced by the mapper step and attach to the generic dict
        files = [
            os.path.join(tmp_dir, f)
            for f in os.listdir(tmp_dir)
            if f.endswith(("fmp", "edge", "log"))
        ]

        for file in files:
            try:
                with open(file, "r") as f:
                    data = f.read()
            except UnicodeDecodeError:
                with open(file, "rb") as f:
                    data = f.read()
            self._add_data_to_generic(file, data)

    def execute(self):
        # run the job in a temporary directory
        tmp_dir = self._make_tmpdir()

        self._write_input_files(tmp_dir)
        self._concatenate_pv_files(tmp_dir)
        io_dict = {
            "": os.path.join(tmp_dir, _SFE.STRUCTCAT_MAEGZ_OUTFILE),
            "-o": _SFE.FEP_MAPPER_OUTPUT,
        }
        arguments = self._parse_arguments(io_dict=io_dict)
        self._apply_token_guard()  # need to implement for reliability
        self._logger.log("Optimising perturbation map", _LE.DEBUG)
        self._backend_executor.execute(
            command=_FE.FEP_MAPPER, arguments=arguments, check=True, location=tmp_dir
        )
        assert os.path.isfile(os.path.join(tmp_dir, "out.fmp"))
        self._logger.log(
            f"Successfully executed fep_mapper in directory {tmp_dir}.", _LE.DEBUG
        )

        self._logger.log("Analysing the perturbation map.", _LE.DEBUG)
        self._analyse_map(tmp_dir)
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



from typing import List
from pydantic import BaseModel, PrivateAttr
from icolos.core.workflow_steps.step import StepSettingsParameters
from icolos.core.workflow_steps.step import StepBase
from icolos.loggers.steplogger import StepLogger
from icolos.core.workflow_steps.step import (
    StepData,
    StepInputParameters,
    StepWriteoutParameters,
    StepExecutionParameters,
)


class BaseStepConfig(BaseModel):
    """
    Minimal template class for the base config, without the unnecessary stuff that StepBase requires
    """

    step_id: str = None
    work_dir: str = None
    type: str = None
    data: StepData = StepData()
    input: StepInputParameters = StepInputParameters()
    writeout: List[StepWriteoutParameters] = []
    execution: StepExecutionParameters = StepExecutionParameters()
    settings: StepSettingsParameters = StepSettingsParameters()

    def as_dict(self):
        return {
            "step_id": self.step_id,
            "type": self.type,
            "execution": self.execution,
            "settings": self.settings,
            "work_dir": self.work_dir,
            "data": self.data,
            "input": self.input,
            "writeout": self.writeout,
        }


class FlowControlBase(BaseModel):
    # List of steps to be iterated over, each set needs their inputs chained together
    base_config: List[BaseStepConfig] = None
    initialized_steps: List[StepBase] = None
    _logger = PrivateAttr()

    def __init__(self, **data) -> None:
        super().__init__(**data)
        self._logger = StepLogger()



import os
from shlex import quote
from icolos.loggers.steplogger import StepLogger
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.execute_external.execute import ExecutorBase
from icolos.utils.enums.program_parameters import SlurmEnum
import subprocess
from typing import List
import time
from tempfile import mkstemp
import numpy as np

_SE = SlurmEnum()
logger = StepLogger()
_LE = LoggingConfigEnum()


class SlurmExecutor(ExecutorBase):
    """For execution of batch jobs to a Slurm cluster."""

    def __init__(
        self,
        cores: int,
        partition: str,
        time: str,
        mem: str,
        tasks: str,
        modules: List,
        other_args: dict,
        gres: str,
        additional_lines: List,
        prefix_execution=None,
        binary_location=None,
    ):
        super().__init__(prefix_execution=None, binary_location=None)

        self.cores = cores
        self.partition = partition
        self.time = time
        self.mem = mem
        self.tasks = tasks
        self.modules = modules
        self.other_args = other_args
        self.gres = gres
        self.additional_lines = additional_lines
        self._script_prefix_execution = prefix_execution
        self._script_binary_location = binary_location

        # check if the machine can reach slurm
        self.slurm_available = self.is_available()
        if not self.slurm_available:
            logger.log(
                "Warning - Slurm was not found, jobs will be run locally!",
                _LE.WARNING,
            )

    def execute(
        self,
        command: str = None,
        arguments: list = None,
        check: bool = True,
        location=None,
        pipe_input=None,
        tmpfile: str = None,
        wait: bool = True,
    ):
        """
        Creates and executes the batch script using the provided resource requirements
        If a path to an existing batch script has not been passed via tmpfile,  it is created
        Attempts to sbatch the jobscript, falling back on bash to provide compatibility with workstations (in this case #SLURM lines are ignored, and the execution becomes blocking)
        """
        if tmpfile is None:
            tmpfile = self.prepare_batch_script(
                command, arguments, pipe_input, location
            )

        if self.slurm_available:
            launch_command = f"sbatch {tmpfile} --no-requeue"
        else:

            launch_command = f"bash {tmpfile}"

        # execute the batch script
        result = super().execute(
            # do not enforce checking here,
            command=launch_command,
            arguments=[],
            location=location,
            check=False,
        )
        if result.returncode != 0:
            # something has gone wrong with submitting the slurm script
            logger.log(
                f"Batch script submission failed with exit code {result.returncode}, error was {result.stderr}",
                _LE.WARNING,
            )
        # either monitor the job id, or resort to parsing the log file
        if self.is_available():
            job_id = result.stdout.split()[-1]
            if wait is False:
                return job_id
            state = self._wait_for_job_completion(job_id=job_id)
        # if using local resources, bash call is blocking, no need to monitor, just wait for result to return
        else:
            state = _SE.COMPLETED if result.returncode == 0 else _SE.FAILED

        # check the result from slurm
        if check == True:
            if state != _SE.COMPLETED:
                raise subprocess.SubprocessError(
                    f"Subprocess returned non-zero exit status:\n{launch_command}\n Status:\n{state}"
                )
        return state

    def prepare_batch_script(
        self,
        command: str,
        arguments: List,
        pipe_input: str = None,
        location: str = None,
    ) -> str:
        """Generate the batch script for a specific job and write to disk

        :param str command: command to be executed
        :param List arguments: List of arguments to be appended to the command
        :param str pipe_input: string to be piped to the program being executed, defaults to None
        :param str location: directory where batch script will be written, defaults to None
        :return str : path to the batch script
        """
        batch_script = self._construct_slurm_header()
        command = self._prepare_command(command, arguments, pipe_input)
        if isinstance(command, str):
            command = [command]
        for cmd in command:
            batch_script.append(cmd)
        _, tmpfile = mkstemp(dir=location, suffix=".sh")
        with open(tmpfile, "w") as f:
            for line in batch_script:
                f.write(line)
                f.write("\n")

        return tmpfile

    def is_available(self):
        command = "sbatch -h"
        result = super().execute(command=command, arguments=[], check=False)
        if any(["Usage: sbatch" in l for l in result.stdout.split("\n")]):
            return True
        return False

    def _prepare_command(
        self, command: str, arguments: List, pipe_input: str = None
    ) -> str:
        arguments = [quote(str(arg)) for arg in arguments]

        # allow for piped input to be passed to binaries
        if pipe_input is not None:
            # pipe_input = self._parse_pipe_input(pipe_input)
            command = pipe_input + " | " + command

        # check, if command (binary) is to be found at a specific location (rather than in $PATH)
        if self._script_binary_location is not None:
            command = os.path.join(self._script_binary_location, command)

        # check, if the something needs to be added before the execution of the "rDock" command
        if self._prefix_execution is not None:
            command = self._script_prefix_execution + " && " + command

        # execute; if "location" is set, change to this directory and execute there
        complete_command = command + " " + " ".join(str(e) for e in arguments)
        complete_command = [complete_command.replace("'", "")]
        return " ".join(complete_command)

    def _wait_for_job_completion(self, job_id: str):
        completed = False
        state = None
        logger.log(f"Monitoring slurm job {job_id}", _LE.DEBUG)
        while completed is False:
            state = self._check_job_status(job_id)
            if state in [_SE.PENDING, _SE.RUNNING, None]:
                time.sleep(60)
                continue
            elif state in (_SE.COMPLETED, _SE.FAILED, _SE.CANCELLED):
                completed = True
        return state

    def _tail_log_file(
        self,
        location: str,
        completed_line: str = "Finished mdrun",
        failed_line: str = "Fatal error",
    ):
        # TODO, this is not very robust at the moment!
        completed = False
        state = None
        while not completed:
            try:
                with open(os.path.join(location, "md.log"), "r") as f:
                    lines = f.readlines()
                completed = any([completed_line in l for l in lines])
                state = _SE.COMPLETED
                failed = any([failed_line in l for l in lines])
                if failed:
                    state = _SE.FAILED
                    for line in lines[-40:]:
                        print(line)
                    completed = True
            except FileNotFoundError:
                logger.log("log file not found, sleeping", _LE.DEBUG)
                time.sleep(10)
        return state

    def _check_job_status(self, job_id):
        """
        Monitor the status of a previously submitted job, return the result
        """
        # use the entrypoint included in the Icolos install

        command = f"sacct -j {job_id} --parsable --noheader -a"
        result = subprocess.run(
            command,
            shell=True,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if result.stdout:
            state = result.stdout.split("\n")[0].split("|")[5].split(" ")[0]

        else:
            state = None
        return state

    def _construct_slurm_header(self):
        header = [
            "#!/bin/bash",
            f"#SBATCH --partition={self.partition}",
            f"#SBATCH --time={self.time}",
        ]
        if self.mem is not None:
            header.append(f"#SBATCH --mem={self.mem}")
        if self.cores is not None:
            header.append(f"#SBATCH -c{self.cores}")
        if self.tasks is not None:
            header.append(f"#SBATCH --tasks={self.tasks}")
        if self.gres is not None:
            header.append(f"#SBATCH --gres={self.gres}")
        for key, value in self.other_args.items():
            if len(str(value)) > 0:
                header.append(f"#SBATCH {key}={value}")
            else:
                header.append(f"#SBATCH {key}")

        for module in self.modules:
            header.append(f"module load {module}")

        # add any other specified lines
        for line in self.additional_lines:
            header.append(line)

        return header



from pydantic import BaseModel
from icolos.core.containers.gmx_state import GromacsState
from icolos.core.workflow_steps.step import StepBase
from typing import List
from icolos.utils.enums.step_enums import StepCavExploreEnum

_SFP = StepCavExploreEnum()


class StepFpocketBase(StepBase, BaseModel):
    eps: float = None
    iso_value: int = None
    threshold: float = None
    min_samples: int = None
    format_: str = None

    def __init__(self, **data):
        super().__init__(**data)

    def _write_input_files(self, tmp_dir, topol: GromacsState):
        for file in self.data.generic.get_flattened_files():
            file.write(tmp_dir)
        if topol is not None:
            # handle structure files
            if not self.data.generic.get_files_by_extension("gro"):
                if topol.structures:
                    topol.write_structure(tmp_dir)
            if not self.data.generic.get_files_by_extension("top"):
                if topol.top_lines:
                    topol.write_topol(tmp_dir)
            if not self.data.generic.get_files_by_extension("tpr"):
                if topol.tprs:
                    topol.write_tpr(tmp_dir)
            if not self.data.generic.get_files_by_extension("xtc"):
                if topol.trajectories:
                    topol.write_trajectory(tmp_dir)
            if not self.data.generic.get_files_by_extension("ndx"):
                if topol.ndx:
                    topol.write_ndx(tmp_dir)

    def _parse_arguments(self, flag_dict: dict, args: list = None) -> List:
        arguments = args if args is not None else []
        # first add the settings from the command line
        for key in self.settings.arguments.parameters.keys():
            arguments.append(key)
            arguments.append(str(self.settings.arguments.parameters[key]))
        for flag in self.settings.arguments.flags:
            arguments.append(str(flag))
        for key, value in flag_dict.items():
            # only add defaults if they have not been specified in the json
            if key not in arguments:
                arguments.append(key)
                arguments.append(value)
        return arguments

    def _set_mdpocket_args(self):
        if self.settings.additional is not None:
            keys = self.settings.additional.keys()

            self.eps = self.settings.additional[_SFP.EPS] if _SFP.EPS in keys else 3
            self.iso_value = (
                self.settings.additional[_SFP.ISO_VALUE]
                if _SFP.ISO_VALUE in keys
                else 0.5
            )
            self.threshold = (
                self.settings.additional[_SFP.THRESHOLD]
                if _SFP.THRESHOLD in keys
                else 20.0
            )
            self.min_samples = (
                self.settings.additional[_SFP.MIN_SAMPLES]
                if _SFP.MIN_SAMPLES in keys
                else 25
            )
            if _SFP.TRAJ_TYPE in keys:
                if self.settings.additional[_SFP.TRAJ_TYPE].lower() == "gromacs":
                    self.format_ = "xtc"
                elif self.settings.additional[_SFP.TRAJ_TYPE].lower() == "desmond":
                    self.format_ = "dtr"
                else:
                    raise ValueError(
                        "Only Desmond and GROMACS trajectory types are supported"
                    )
            else:
                raise ValueError("Trajectory format was not specified!")



from icolos.core.flow_control.iterator import StepIterator
from icolos.core.workflow_steps.active_learning.virtual_screening import (
    StepActiveLearning,
)
from icolos.utils.enums.step_enums import StepBaseEnum

_SBE = StepBaseEnum


class FlowControlInitializationEnum:
    # These steps are responsible for initializing other steps as part of their execution
    # Keep these separate to the main pool of steps to avoid circular imports

    FLOW_CONTROL_INIT_DICT = {
        _SBE.STEP_ITERATOR: StepIterator,
        _SBE.STEP_ACTIVE_LEARNING: StepActiveLearning,
    }
